!****************************************************************************
!
!  MODULE:  hm_utils
!
!  PURPOSE: contains useful functions 
!
!****************************************************************************

module hm_utils
use nrtype
implicit none

contains



! **********************************************************************

SUBROUTINE unit_matrix(mat)
real(DP), DIMENSION(:,:), INTENT(OUT) :: mat

! Sets the diagonal components of mat to unity, all other components to zero.
! When mat is square, this will be the unit matrix; otherwise, a unit matrix
! with appended rows or columns of zeros.

integer(I4B):: i,n
n=min(size(mat,1),size(mat,2))

mat(:,:)=0.0d0
do i=1,n
	mat(i,i)=1.0d0
end do

END SUBROUTINE unit_matrix



! **********************************************************************

function npdf(L,Z,CHOL)
! computes scalar value of multivariate normal density function
integer(I4B)::L
real(DP)::npdf,DET,Z(L),CHOL(L,L),COV_INV(L,L),UNITM(L,L)

DET=dettriang(L,CHOL)
COV_INV=matmul(transpose(CHOL),CHOL)
! invert COV_INV
call unit_matrix(UNITM)
call gaussj(COV_INV,UNITM)

npdf=(2.0d0*pi_d)**(-L/2.0d0)/DET*exp(-.5d0*sum(Z*matmul(COV_INV,Z)))

end function npdf



! **********************************************************************

function lognpdf(Z,MU,SIG)
! computes scalar value of UNIVARIATE log normal density function 
! where the underlying normal density is MU,SIG
integer(I4B)::L
real(DP)::lognpdf,Z,MU,SIG 

lognpdf=((2.0d0*pi_d)**.5d0 *SIG*Z)**(-1.0d0)*exp(-.5d0*((log(Z)-MU)/SIG)**2.0d0)

end function lognpdf



! **********************************************************************

function logitnpdf(Z,MU,SIG)
! computes scalar value of UNIVARIATE ``logit'' normal density function 
! where the underlying normal density is MU,SIG
integer(I4B)::L
real(DP)::logitnpdf,Z,MU,SIG

logitnpdf=((2.0d0*pi_d)**.5d0 *SIG*Z*(1.0d0-Z))**(-1.0d0)*exp(-.5d0*((log(Z/(1.0d0-Z))-MU)/SIG)**2.0d0)

end function logitnpdf



! **********************************************************************

function logit2npdf(Z,MU,SIG)
! computes scalar value of UNIVARIATE ``logit[-1,1]'' normal density 
! function where the underlying normal density is MU,SIG
integer(I4B)::L
real(DP)::logit2npdf,Z,MU,SIG

logit2npdf=((2.0d0*pi_d)**.5d0 *SIG*(1.0d0-Z**2.0d0))**(-1.0d0)* &
           exp(-.5d0*((log((1.0d0+Z)/(1.0d0-Z))-MU)/SIG)**2.0d0)

end function logit2npdf



! **********************************************************************

function ncdf(x)
use nrtype
implicit none
real(dp),intent(in)::x
real(dp)::ncdf

ncdf = 1.0d0-erfcc_s(x/sqrt(2.0d0))/2.0d0

end function ncdf




! **********************************************************************

function ncdfinv(px)
use nrtype
implicit none
real(dp),intent(in)::px
real(dp)::ncdfinv

ncdfinv=zbrent(ncdfinv_hlp,-1.0d5,1.0d5,1.0d-15,px)

end function ncdfinv




! **********************************************************************

function ncdfinv_hlp(x,px)
use nrtype
implicit none
real(dp),intent(in)::x,px
real(dp)::ncdfinv_hlp

ncdfinv_hlp = ncdf(x)-px

end function ncdfinv_hlp




! **********************************************************************

function logistcdf(Z)
! computes scalar logistic cdf 
real(DP) logistcdf,Z

logistcdf=1.0d0/(1.0d0+exp(-Z))

end function logistcdf



! **********************************************************************

function logistpdf(Z)
! computes scalar logistic pdf 
real(DP) logistpdf,Z

logistpdf=logistcdf(Z)*(1.0d0-logistcdf(Z))

end function logistpdf



! **********************************************************************

function logistinv(Z)
! computes scalar inverse of logistic distribution function
real(DP) logistinv,Z

logistinv=-log((1.0d0-Z)/Z)

end function logistinv



! **********************************************************************

subroutine corr2chol(L,CORR,CHOL)
! computes cholesky decomposition from a correlation matrix
integer(I4B), intent(in):: L
real(DP),intent(in) ::CORR(L,L)
real(DP)            :: COV(L,L)
real(DP),intent(out)::CHOL(L,L)
integer(I4B) i,j
logical(lgt):: pd_flag

do i=1,L
  do j=1,i
    if (j .NE. i) then
	  COV(j,i)=CORR(j,i)*CORR(i,i)*CORR(j,j)
	  ! this assumes that CORR is upper triangular or symmetric
	  COV(i,j)=COV(j,i)
	endif
  enddo
enddo

do i=1,L
  COV(i,i)=(CORR(i,i))**2.0d0
enddo

call choldc(L,COV,CHOL,pd_flag)

end subroutine corr2chol



! **********************************************************************

subroutine corr2cov(L,CORR,COV)
! computes covariance matrix from a correlation matrix
integer(I4B), intent(in):: L
real(DP),intent(in) ::CORR(L,L)
real(DP),intent(out):: COV(L,L)
integer(I4B) i,j

do i=1,L
  do j=1,i
    if (j .NE. i) then
	  COV(j,i)=CORR(j,i)*CORR(i,i)*CORR(j,j)
	  ! this assumes that CORR is upper triangular or symmetric
	  COV(i,j)=COV(j,i)
	endif
  enddo
enddo

do i=1,L
  COV(i,i)=(CORR(i,i))**2.0d0
enddo

end subroutine corr2cov



! **********************************************************************

subroutine chol2corr(L,CHOL,CORR)
! computes correlation matrix from an upper triangular cholesky matrix
integer(I4B), intent(in):: L
real(DP),intent(in) ::CHOL(L,L)
real(DP)            :: COV(L,L)
real(DP),intent(out)::CORR(L,L)
integer(I4B) i,j,k

COV=matmul(transpose(CHOL),CHOL)

do i=1,L
  CORR(i,i)=sqrt(COV(i,i))
enddo

do i=1,L
  do j=1,i
    if (j .NE. i) then
	  CORR(i,j)= COV(i,j)/CORR(i,i)/CORR(j,j)
	  CORR(j,i)=CORR(i,j)
	endif
  enddo
enddo


end subroutine chol2corr



! **********************************************************************

subroutine cov2corr(L,COV,CORR)
! computes correlation matrix from a covariance matrix
integer(I4B), intent(in):: L
real(DP),intent(in) ::COV(L,L)
real(DP),intent(out)::CORR(L,L)
integer(I4B) i,j,k

do i=1,L
  if (COV(i,i) .GE. 0.0d0) then
    CORR(i,i)=sqrt(COV(i,i))
  else
    CORR(i,i)=0.0d0
  endif
enddo

do i=1,L
  do j=1,i
    if (j .NE. i) then
	  CORR(i,j)=COV(i,j)/CORR(i,i)/CORR(j,j)
	  CORR(j,i)=CORR(i,j)
	endif
  enddo
enddo


end subroutine cov2corr









! **********************************************************************

subroutine choldc(L,COVA,CHOLE,flag)
USE nrtype
IMPLICIT NONE
INTEGER(I4B),INTENT(IN):: L
REAL(DP), DIMENSION(L,L), INTENT(IN)  :: COVA
REAL(DP), DIMENSION(L,L), INTENT(OUT) :: CHOLE
LOGICAL(LGT),INTENT(OUT):: flag
!Given an L  L positive-definite symmetric matrix COV, this routine constructs its Cholesky
!decomposition, COV = CHOL^T  CHOL . On input, only the upper triangle of COV need be given; 
!it is not modified. The Cholesky factor is returned in the upper triangle of CHOL
INTEGER(I4B) :: i,j
REAL(DP) :: summ

CHOLE=0.0d0
do i=1,L
  summ=COVA(i,i)-dot_product(COVA(i,1:i-1),COVA(i,1:i-1))
  if (summ <= 0.0) then
    flag=.FALSE.
    return
  endif
  CHOLE(i,i)=sqrt(summ)
  CHOLE(i+1:L,i)=(COVA(i,i+1:L)-matmul(CHOLE(i+1:L,1:i-1),COVA(i,1:i-1)))/CHOLE(i,i)
end do

CHOLE=transpose(CHOLE)
flag=.TRUE.

END SUBROUTINE choldc



! **********************************************************************

function dettriang(K,Z)
! computes determinant of a triangular matrix
integer(I4B):: K,j
real(DP)::dettriang,Z(K,K)
dettriang=Z(1,1)
if (K>1) then 
	do j=2,K
		dettriang=dettriang*Z(j,j)
	enddo
endif
end function dettriang



! **************************************************************

subroutine vec2triangmat(L,vec,triangmat)
! converts vector with dimension (L*(L+1)/2) into 
! LxL upper triangular matrix
integer(I4B),intent(in)::L
real(DP),intent(in) ::vec((L*(L+1))/2)
real(DP),intent(out)::triangmat(L,L)
integer(I4B)::i1,i2,i3

triangmat=0.0d0
i1=1
do i2=1,L
  i3=1
  do while (i3 .LE. i2)
    triangmat(i3,i2)=vec(i1)
    i1=i1+1
    i3=i3+1
  enddo
enddo

end subroutine vec2triangmat



! **************************************************************

subroutine triangmat2vec(L,triangmat,vec)
! converts LxL upper triangular matrix
! into (L*(L+1)/2) vector 
integer(I4B),intent(in)::L
real(DP),intent(in)::triangmat(L,L)
real(DP),intent(out) ::vec((L*(L+1))/2)
integer(I4B)::i1,i2,i3

vec=0.0d0
i1=1
do i2=1,L
  i3=1
  do while (i3 .LE. i2)
    vec(i1)=triangmat(i3,i2)
    i1=i1+1
    i3=i3+1
  enddo
enddo

end subroutine triangmat2vec




! **********************************************************************

subroutine partition_cov(Lpart1,L,loc1,loc2,origmat,matA,matAB,matB)
! partitions a covariance matrix of size L in three matrices:
! L1xL1 partition A (L1<=2)
! L1xL2 partition AB
! L2xL2 partition B
USE nrtype 
USE nrutil, ONLY : nrerror
integer(I4B),intent(in)::Lpart1,L
real(DP),intent(in):: origmat(L,L)
real(DP),intent(out)::matA(Lpart1,Lpart1),matAB(Lpart1,L-Lpart1),matB(L-Lpart1,L-Lpart1)
integer(I4B)::loc1,loc2,j,i,jA,iA,jB,iB,jAB,iAB



if (Lpart1==1) then

  iAB=1; 
  jAB=1; 
  iB=1; 
  jB=1
  do i=1,L
    do j=1,L
      if (i==loc1 .AND. j==loc1) then; matA(1,1)   =origmat(i,j); endif
      if (i==loc1 .AND. j <loc1) then; matAB(1,jAB)=origmat(i,j); jAB=jAB+1; endif
      if (i==loc1 .AND. j >loc1) then; matAB(1,jAB)=origmat(i,j); jAB=jAB+1; endif
      if (i/=loc1 .AND. j <loc1) then; matB(iB,jB) =origmat(i,j); matB(jB,iB) =origmat(i,j); jB=jB+1; endif
      if (i/=loc1 .AND. j >loc1) then; matB(iB,jB) =origmat(i,j); matB(jB,iB) =origmat(i,j); jB=jB+1; endif
    enddo
    if (i/=loc1) iB=iB+1
    jB=1; 
  enddo

elseif (Lpart1==2) then

  if (loc1==loc2) call nrerror('location parameters must differ in partition_cov')
  if (loc1>loc2)  then; j=loc1; loc1=loc2; loc2=j; endif


  iAB=1; 
  jAB=1; 
  iB=1; 
  jB=1
  do i=1,L
    do j=1,L
      if (i==loc1 .AND. j==loc1) then; matA(1,1)   =origmat(i,j); endif
      if (i==loc1 .AND. j==loc2) then; matA(1,2)   =origmat(i,j); matA(2,1)=origmat(i,j); endif
      if (i==loc2 .AND. j==loc2) then; matA(2,2)   =origmat(i,j); endif
      if (i==loc1 .AND. j <loc1 .AND. j/=loc2) then; matAB(1,jAB)=origmat(i,j); jAB=jAB+1; endif
      if (i==loc1 .AND. j >loc1 .AND. j/=loc2) then; matAB(1,jAB)=origmat(i,j); jAB=jAB+1; endif
      if (i==loc2 .AND. j <loc2 .AND. j/=loc1) then; matAB(2,jAB)=origmat(i,j); jAB=jAB+1; endif
      if (i==loc2 .AND. j >loc2 .AND. j/=loc1) then; matAB(2,jAB)=origmat(i,j); jAB=jAB+1; endif
      if (i/=loc1 .AND. i/=loc2 .AND. j <loc1 .AND. j <loc2 ) then; matB(iB,jB) =origmat(i,j); matB(jB,iB)=origmat(i,j); jB=jB+1; endif
      if (i/=loc1 .AND. i/=loc2 .AND. j >loc1 .AND. j <loc2 ) then; matB(iB,jB) =origmat(i,j); matB(jB,iB)=origmat(i,j); jB=jB+1; endif
      if (i/=loc1 .AND. i/=loc2 .AND. j >loc1 .AND. j >loc2 ) then; matB(iB,jB) =origmat(i,j); matB(jB,iB)=origmat(i,j); jB=jB+1; endif
    enddo
    if (i/=loc1 .AND. i/=loc2) iB=iB+1
    jB=1;
	jAB=1;
  enddo

else
  call nrerror('dimension of partition A too large for partition_cov')
endif 


end subroutine partition_cov

! **********************************************************************

SUBROUTINE gaussj(a,b)
USE nrtype 
USE nrutil, ONLY : assert_eq,nrerror,outerand,outerprod,swap
IMPLICIT NONE
REAL(SP), DIMENSION(:,:), INTENT(INOUT) :: a,b
! Linear equation solution by Gauss-Jordan elimination, equation (2.1.1). a is an NN input
! coefficient matrix. b is an NM input matrix containing M right-hand-side vectors. On
! output, a is replaced by its matrix inverse, and b is replaced by the corresponding set of
! solution vectors.
INTEGER(I4B), DIMENSION(size(a,1)) :: ipiv,indxr,indxc
! These arrays are used for bookkeeping on the pivoting.
LOGICAL(LGT), DIMENSION(size(a,1)) :: lpiv
REAL(SP) :: pivinv
REAL(SP), DIMENSION(size(a,1)) :: dumc
INTEGER(I4B), TARGET :: irc(2)
INTEGER(I4B) :: i,l,n
INTEGER(I4B), POINTER :: irow,icol

n=assert_eq(size(a,1),size(a,2),size(b,1),'gaussj')
irow => irc(1)
icol => irc(2)
ipiv=0
do i=1,n ! Main loop over columns to be reduced.
  lpiv = (ipiv == 0) ! Begin search for a pivot element.
  irc=maxloc(abs(a),outerand(lpiv,lpiv))
  ipiv(icol)=ipiv(icol)+1
  if (ipiv(icol) > 1) call nrerror('gaussj: singular matrix (1)')
  ! We now have the pivot element, so we interchange rows, if needed, to put the pivot
  ! element on the diagonal. The columns are not physically interchanged, only relabeled:
  ! indxc(i), the column of the ith pivot element, is the ith column that is reduced, while
  ! indxr(i) is the row in which that pivot element was originally located. If indxr(i) .=
  ! indxc(i) there is an implied column interchange. With this form of bookkeeping, the
  ! solution bs will end up in the correct order, and the inverse matrix will be scrambled
  ! by columns.
  if (irow /= icol) then
    call swap(a(irow,:),a(icol,:))
    call swap(b(irow,:),b(icol,:))
  end if
  indxr(i)=irow        ! We are now ready to divide the pivot row by the pivot
  indxc(i)=icol        ! element, located at irow and icol. 
  if (a(icol,icol) == 0.0) then
    call nrerror('gaussj: singular matrix (2)')
  endif
  pivinv=1.0d0/a(icol,icol)
  a(icol,icol)=1.0
  a(icol,:)=a(icol,:)*pivinv
  b(icol,:)=b(icol,:)*pivinv
  dumc=a(:,icol) ! Next, we reduce the rows, except for the pivot one, of course. 
  a(:,icol)=0.0
  a(icol,icol)=pivinv
  a(1:icol-1,:)=a(1:icol-1,:)-outerprod(dumc(1:icol-1),a(icol,:))
  b(1:icol-1,:)=b(1:icol-1,:)-outerprod(dumc(1:icol-1),b(icol,:))
  a(icol+1:,:)=a(icol+1:,:)-outerprod(dumc(icol+1:),a(icol,:))
  b(icol+1:,:)=b(icol+1:,:)-outerprod(dumc(icol+1:),b(icol,:))
end do
! It only remains to unscramble the solution in view of the column interchanges. We do this
! by interchanging pairs of columns in the reverse order that the permutation was built up.
do l=n,1,-1
  call swap(a(:,indxr(l)),a(:,indxc(l)))
end do
END SUBROUTINE gaussj




! **********************************************************************

FUNCTION erfcc_s(x)
USE nrtype; USE nrutil, ONLY : poly_dd
IMPLICIT NONE
REAL(DP), INTENT(IN) :: x
REAL(DP) :: erfcc_s
! Returns the complementary error function erfc(x) with fractional error everywhere less than 1.2  10-7.
REAL(DP) :: t,z
REAL(DP), DIMENSION(10) :: coef = (/-1.26551223d0,1.00002368d0,&
0.37409196d0,0.09678418d0,-0.18628806d0,0.27886807d0,&
-1.13520398d0,1.48851587d0,-0.82215223d0,0.17087277d0/)
z=abs(x)
t=1.0d0/(1.0d0+0.5d0*z)
erfcc_s=t*exp(-z*z+poly_dd(t,coef))
if (x < 0.0) erfcc_s=2.0d0-erfcc_s
END FUNCTION erfcc_s



! **********************************************************************

FUNCTION zbrent(func,x1,x2,tol,funcinput)
use nrtype
REAL(DP)::zbrent,tol,x1,x2,func,funcinput
EXTERNAL:: func
integer(I4B),parameter::ITMAX=1000
real(DP),parameter::EPS=tiny(tol)

! Using Brent's method, find the root of a function func known to lie between x1 and x2.
! The root, returned as zbrent, will be refined until its accuracy is tol.
! Parameters: Maximum allowed number of iterations, and machine floating-point precision.

INTEGER(I4B):: iter
REAL(DP):: a,b,c,d,e,fa,fb,fc,p,q,r,s,tol1,xm


a=x1
b=x2
fa=func(a,funcinput)
fb=func(b,funcinput)
if((fa.gt.0..and.fb.gt.0.).or.(fa.lt.0..and.fb.lt.0.)) pause 'root must be bracketed for zbrent'
c=b
fc=fb
do iter=1,ITMAX
  if((fb.gt.0..and.fc.gt.0.).or.(fb.lt.0..and.fc.lt.0.)) then
    c=a !Rename a, b, c and adjust bounding interval d.
    fc=fa
    d=b-a
    e=d
  endif
  if(abs(fc).lt.abs(fb)) then
    a=b
    b=c
    c=a
    fa=fb
    fb=fc
    fc=fa
  endif
  tol1=2.*EPS*abs(b)+0.5*tol  ! Convergence check.
  xm=.5*(c-b)
  if(abs(xm).le.tol1 .or. fb.eq.0.)then
    zbrent=b
    return
  endif
  if(abs(e).ge.tol1 .and. abs(fa).gt.abs(fb)) then
    s=fb/fa ! Attempt inverse quadratic interpolation.
    if(a.eq.c) then
      p=2.*xm*s
      q=1.-s
    else
      q=fa/fc
      r=fb/fc
      p=s*(2.*xm*q*(q-r)-(b-a)*(r-1.))
      q=(q-1.)*(r-1.)*(s-1.)
    endif
    if(p.gt.0.) q=-q ! Check whether in bounds.
      p=abs(p)
      if(2.*p .lt. min(3.*xm*q-abs(tol1*q),abs(e*q))) then
        e=d ! Accept interpolation.
        d=p/q
      else
        d=xm ! Interpolation failed, use bisection.
        e=d
      endif
    else  ! Bounds decreasing too slowly, use bisection.
      d=xm
      e=d
    endif
      a=b ! Move last best guess to a.
      fa=fb
      if(abs(d) .gt. tol1) then ! Evaluate new trial root.
        b=b+d
      else      
        b=b+sign(tol1,xm)
      endif
    fb=func(b,funcinput)
  enddo
  pause 'zbrent exceeding maximum iterations'
  zbrent=b
  return
END function zbrent



! **********************************************************************

subroutine printarray(dim1,dim2,ARRAY,file)
integer(I4B),intent(in)::dim1,dim2,file
real(DP),intent(in)::ARRAY(dim1,dim2)

write(file,*) ' '

select case (dim2)
case  (1) 
  write (file,FMT= 1) transpose(ARRAY)
case  (2) 
  write (file,FMT= 2) transpose(ARRAY)
case  (3) 
  write (file,FMT= 3) transpose(ARRAY)
case  (4) 
  write (file,FMT= 4) transpose(ARRAY)
case  (5) 
  write (file,FMT= 5) transpose(ARRAY)
case  (6) 
  write (file,FMT= 6) transpose(ARRAY)
case  (7) 
  write (file,FMT= 7) transpose(ARRAY)
case  (8) 
  write (file,FMT= 8) transpose(ARRAY)
case  (9) 
  write (file,FMT= 9) transpose(ARRAY)
case (10) 
  write (file,FMT=10) transpose(ARRAY)
case (11) 
  write (file,FMT=11) transpose(ARRAY)
case (12) 
  write (file,FMT=12) transpose(ARRAY)
case(13) 
  write (file,FMT=13) transpose(ARRAY)
case(14) 
  write (file,FMT=14) transpose(ARRAY)
case(15) 
  write (file,FMT=15) transpose(ARRAY)
case(16) 
  write (file,FMT=16) transpose(ARRAY)
case(17) 
  write (file,FMT=17) transpose(ARRAY)
case(18) 
  write (file,FMT=18) transpose(ARRAY)
case(19) 
  write (file,FMT=19) transpose(ARRAY)
case(20) 
  write (file,FMT=20) transpose(ARRAY)
case(21) 
  write (file,FMT=21) transpose(ARRAY)
case(22) 
  write (file,FMT=22) transpose(ARRAY)
case(23) 
  write (file,FMT=23) transpose(ARRAY)
case(24) 
  write (file,FMT=24) transpose(ARRAY)
case(25) 
  write (file,FMT=25) transpose(ARRAY)
case(26) 
  write (file,FMT=26) transpose(ARRAY)
case(27) 
  write (file,FMT=27) transpose(ARRAY)
case(28) 
  write (file,FMT=28) transpose(ARRAY)
case(29) 
  write (file,FMT=29) transpose(ARRAY)
case(30) 
  write (file,FMT=30) transpose(ARRAY)
case(31) 
  write (file,FMT=31) transpose(ARRAY)
case(32) 
  write (file,FMT=32) transpose(ARRAY)
case(33) 
  write (file,FMT=33) transpose(ARRAY)
case(34) 
  write (file,FMT=34) transpose(ARRAY)
case(35) 
  write (file,FMT=35) transpose(ARRAY)
case default
	print*,'dim2 too large or specified incorrectly for subroutine printarray'
	pause		
end select

write(file,*) ' '

 1 format( 1(1X,E15.8,1X))
 2 format( 2(1X,E15.8,1X))
 3 format( 3(1X,E15.8,1X))
 4 format( 4(1X,E15.8,1X))
 5 format( 5(1X,E15.8,1X))
 6 format( 6(1X,E15.8,1X))
 7 format( 7(1X,E15.8,1X))
 8 format( 8(1X,E15.8,1X))
 9 format( 9(1X,E15.8,1X))
10 format(10(1X,E15.8,1X))
11 format(11(1X,E15.8,1X))
12 format(12(1X,E15.8,1X))
13 format(13(1X,E15.8,1X))
14 format(14(1X,E15.8,1X))
15 format(15(1X,E15.8,1X))
16 format(16(1X,E15.8,1X))
17 format(17(1X,E15.8,1X))
18 format(18(1X,E15.8,1X))
19 format(19(1X,E15.8,1X))
20 format(20(1X,E15.8,1X))
21 format(21(1X,E15.8,1X))
22 format(22(1X,E15.8,1X))
23 format(23(1X,E15.8,1X))
24 format(24(1X,E15.8,1X))
25 format(25(1X,E15.8,1X))
26 format(26(1X,E15.8,1X))
27 format(27(1X,E15.8,1X))
28 format(28(1X,E15.8,1X))
29 format(29(1X,E15.8,1X))
30 format(30(1X,E15.8,1X))
31 format(31(1X,E15.8,1X))
32 format(32(1X,E15.8,1X))
33 format(33(1X,E15.8,1X))
34 format(34(1X,E15.8,1X))
35 format(35(1X,E15.8,1X))

end subroutine printarray


! **********************************************************************

subroutine printarraycsv(dim1,dim2,ARRAY,file)
integer(I4B),intent(in)::dim1,dim2,file
real(DP),intent(in)::ARRAY(dim1,dim2)

write(file,*) ' '

select case (dim2)
case  (1) 
  write (file,FMT= 1) transpose(ARRAY)
case  (2) 
  write (file,FMT= 2) transpose(ARRAY)
case  (3) 
  write (file,FMT= 3) transpose(ARRAY)
case  (4) 
  write (file,FMT= 4) transpose(ARRAY)
case  (5) 
  write (file,FMT= 5) transpose(ARRAY)
case  (6) 
  write (file,FMT= 6) transpose(ARRAY)
case  (7) 
  write (file,FMT= 7) transpose(ARRAY)
case  (8) 
  write (file,FMT= 8) transpose(ARRAY)
case  (9) 
  write (file,FMT= 9) transpose(ARRAY)
case (10) 
  write (file,FMT=10) transpose(ARRAY)
case (11) 
  write (file,FMT=11) transpose(ARRAY)
case (12) 
  write (file,FMT=12) transpose(ARRAY)
case(13) 
  write (file,FMT=13) transpose(ARRAY)
case(14) 
  write (file,FMT=14) transpose(ARRAY)
case(15) 
  write (file,FMT=15) transpose(ARRAY)
case(16) 
  write (file,FMT=16) transpose(ARRAY)
case(17) 
  write (file,FMT=17) transpose(ARRAY)
case(18) 
  write (file,FMT=18) transpose(ARRAY)
case(19) 
  write (file,FMT=19) transpose(ARRAY)
case(20) 
  write (file,FMT=20) transpose(ARRAY)
case(21) 
  write (file,FMT=21) transpose(ARRAY)
case(22) 
  write (file,FMT=22) transpose(ARRAY)
case(23) 
  write (file,FMT=23) transpose(ARRAY)
case(24) 
  write (file,FMT=24) transpose(ARRAY)
case(25) 
  write (file,FMT=25) transpose(ARRAY)
case(26) 
  write (file,FMT=26) transpose(ARRAY)
case(27) 
  write (file,FMT=27) transpose(ARRAY)
case(28) 
  write (file,FMT=28) transpose(ARRAY)
case(29) 
  write (file,FMT=29) transpose(ARRAY)
case(30) 
  write (file,FMT=30) transpose(ARRAY)
case(31) 
  write (file,FMT=31) transpose(ARRAY)
case(32) 
  write (file,FMT=32) transpose(ARRAY)
case(33) 
  write (file,FMT=33) transpose(ARRAY)
case(34) 
  write (file,FMT=34) transpose(ARRAY)
case(35) 
  write (file,FMT=35) transpose(ARRAY)
case default
	print*,'dim2 too large or specified incorrectly for subroutine printarraycsv'
	pause		
end select

write(file,*) ' '

 1 format( 1(1X,E15.8))
 2 format( 2(1X,E15.8,';'))
 3 format( 3(1X,E15.8,';'))
 4 format( 4(1X,E15.8,';',1X))
 5 format( 5(1X,E15.8,';',1X))
 6 format( 6(1X,E15.8,';',1X))
 7 format( 7(1X,E15.8,';',1X))
 8 format( 8(1X,E15.8,';',1X))
 9 format( 9(1X,E15.8,';',1X))
10 format(10(1X,E15.8,';',1X))
11 format(11(1X,E15.8,';',1X))
12 format(12(1X,E15.8,';',1X))
13 format(13(1X,E15.8,';',1X))
14 format(14(1X,E15.8,';',1X))
15 format(15(1X,E15.8,';',1X))
16 format(16(1X,E15.8,';',1X))
17 format(17(1X,E15.8,';',1X))
18 format(18(1X,E15.8,';',1X))
19 format(19(1X,E15.8,';',1X))
20 format(20(1X,E15.8,';',1X))
21 format(21(1X,E15.8,';',1X))
22 format(22(1X,E15.8,';',1X))
23 format(23(1X,E15.8,';',1X))
24 format(24(1X,E15.8,';',1X))
25 format(25(1X,E15.8,';',1X))
26 format(26(1X,E15.8,';',1X))
27 format(27(1X,E15.8,';',1X))
28 format(28(1X,E15.8,';',1X))
29 format(29(1X,E15.8,';',1X))
30 format(30(1X,E15.8,';',1X))
31 format(31(1X,E15.8,';',1X))
32 format(32(1X,E15.8,';',1X))
33 format(33(1X,E15.8,';',1X))
34 format(34(1X,E15.8,';',1X))
35 format(35(1X,E15.8,';',1X))

end subroutine printarraycsv


! **********************************************************************

subroutine printvec(dim1,VEC,file)
integer(I4B),intent(in)::dim1,file
real(DP),intent(in)::VEC(dim1)
real(DP)::ARRAY(dim1,1)
integer::i

do i=1,dim1
  ARRAY(i,1)=VEC(i)
enddo

write(file,*) ' '

write (file,FMT= 1) transpose(ARRAY)

write(file,*) ' '

 1 format( 1(1X,E15.8,1X))

end subroutine printvec


! **********************************************************************

subroutine stepfwd(x,xh)
real(DP),intent(in)::x
real(DP),intent(inout)::xh
real(DP)::temp(1)
integer(I4B)::i


! make sure that step is actually representable exactly
temp(1)=x+xh
call donothing(temp,temp)
xh=temp(1)-x

end subroutine stepfwd



! **********************************************************************

subroutine donothing(temp1,temp2)
use nrtype
implicit none
real(dp)::temp1(:),temp2(:)
end subroutine donothing



! **********************************************************************

FUNCTION dfridr(func,x,h,err)
USE nrtype; USE nrutil, ONLY : assert,geop,iminloc
IMPLICIT NONE
REAL(DP), INTENT(IN) :: x,h
REAL(DP), INTENT(OUT) :: err
REAL(DP) :: dfridr
INTERFACE
  FUNCTION func(x)
  USE nrtype
  IMPLICIT NONE
  REAL(DP), INTENT(IN) :: x
  REAL(DP) :: func
  END FUNCTION func
END INTERFACE
INTEGER(I4B),PARAMETER :: NTAB=10
REAL(DP), PARAMETER :: CON=1.4d0,CON2=CON*CON,BIG=huge(x),SAFE=5.0d0
! Returns the derivative of a function func at a point x by Ridders method of polynomial
! extrapolation. The value h is input as an estimated initial stepsize; it need not be small,
! but rather should be an increment in x over which func changes substantially. An estimate
! of the error in the derivative is returned as err.
! Parameters: Stepsize is decreased by CON at each iteration. Max size of tableau is set by
! NTAB. Return when error is SAFE worse than the best so far.
INTEGER(I4B) :: ierrmin,i,j
REAL(DP) :: hh
REAL(DP), DIMENSION(NTAB-1) :: errt,fac
REAL(DP), DIMENSION(NTAB,NTAB) :: a
call assert(h /= 0.0, 'dfridr arg')
hh=h
a(1,1)=(func(x+hh)-func(x-hh))/(2.0d0*hh)
err=BIG
fac(1:NTAB-1)=geop(CON2,CON2,NTAB-1)
do i=2,NTAB 
  ! Successive columns in the Neville tableau will go to smaller
  ! stepsizes and higher orders of extrapolation.
  hh=hh/CON 
  a(1,i)=(func(x+hh)-func(x-hh))/(2.0d0*hh) ! Try new, smaller stepsize.
  do j=2,i
    ! Compute extrapolations of various orders, requiring no new function evaluations.
    a(j,i)=(a(j-1,i)*fac(j-1)-a(j-1,i-1))/(fac(j-1)-1.0d0)
  end do
  errt(1:i-1)=max(abs(a(2:i,i)-a(1:i-1,i)),abs(a(2:i,i)-a(1:i-1,i-1)))
  ! The error strategy is to compare each new extrapolation to one order lower, both at the
  ! present stepsize and the previous one.
  ierrmin=iminloc(errt(1:i-1))
  if (errt(ierrmin) <= err) then 
  ! If error is decreased, save the improved answer.
    err=errt(ierrmin) 
    dfridr=a(1+ierrmin,i)
  end if
  if (abs(a(i,i)-a(i-1,i-1)) >= SAFE*err) RETURN
  ! If higher order is worse by a significant factor SAFE, then quit early.
end do
END FUNCTION dfridr



! **********************************************************************

end module hm_utils