!****************************************************************************
!
!  SUBROUTINE:	logl_all
!
!  PURPOSE:		core function: contains all estimation steps
!
!****************************************************************************

subroutine logl_all(KK,COEFFS,nf,LOGL_VALUE,uip,urp,ufp)
use controlpars
use hm_utils
implicit none




! parameters in function definition
integer(I4B),intent(in)::KK      ! number of parameters, restated because of interface (same as K)
real(DP),intent(in)::COEFFS(KK)  ! coefficient vector
real(DP)::LOGL_VALUE             ! overall objective function
integer(I4B),intent(in)::nf,uip  ! number of function evaluations / unused dummy argument
real(DP),intent(in)::urp,ufp     ! unused dummy arguments
real(DP)::NODES(L_COV_RC,NDRAWS)




! additional parameters 
real(DP)::LIKEL_ind_global(N)    ! vector of individual likelihood contributions
real(DP)::LIKEL_orig(N)   ! vector of original individual likelihood contributions, needed for gradient calcs
logical(lgt)::NEWITER     ! printing control parameter
integer(I4B)::NCALCS,NC,R ! number of calculations / counters 
real(DP)::ridd_err        ! bound on error in gradient calculation




! individual-level preference parameter averages and distributions
real(DP)::GAM_AVG(N),LAM_AVG(N),RHO_AVG(N),TAU_AVG(N),OMG_AVG(N)
real(DP)::GAM_AVG_global(N),LAM_AVG_global(N),RHO_AVG_global(N),TAU_AVG_global(N),OMG_AVG_global(N)





! parallel-computing related business
integer(I4B):: local_N            ! length of local loop
integer(I4B):: local_ind_beg      ! Starting point of loop for my process.
integer(I4B):: local_ind_end      ! Endpoint of loop my process.
real(DP):: LIKEL_ind(N)           ! Local version of LIKEL_ind_global
real(DP):: local_GRADIENT_ind(N)  ! Local version of GRADIENT_ind
real(DP)::  temp_GRADIENT_ind(N)  ! Local version of GRADIENT_ind used to receive other processes' calculations
integer(I4B):: source             ! Process sending integal.
integer(I4B),parameter:: dest=0   ! All messages go to 0.
integer(I4B),parameter:: tag1=11
integer(I4B),parameter:: tag2=22
integer(I4B),parameter:: tag3=33
integer(I4B),parameter:: tag4=44
integer(I4B),parameter:: tag5=55
integer(I4B),parameter:: tag6=66
integer(I4B),parameter:: tag7=77
integer(I4B),parameter:: tag8=88




! initialise some values

! set number of calculations if gradient is to be evaluated 
NCALCS=1
if (CALCGRAD) then
  NCALCS=K
endif 
! set gradient calculation error to zero
ridders_err=0.0d0
! printing control: check whether new iteration starts
NEWITER=.FALSE.
if (funccounter<nf) then
  NEWITER=.TRUE.
  funccounter=nf
endif 
! set individual-level vectors to zero
GAM_AVG=0.0d0;LAM_AVG=0.0d0;RHO_AVG=0.0d0;TAU_AVG=0.0d0;OMG_AVG=0.0d0




! some MPI-related business
if (my_rank .EQ. 0) then
  ! distribute coefficients to other machines
  call MPI_BCAST(PARALLITER,1,MPI_LOGICAL,         dest,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(CALCGRAD  ,1,MPI_LOGICAL,         dest,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(CALCIPAR  ,1,MPI_LOGICAL,         dest,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(    COEFFS,K,MPI_DOUBLE_PRECISION,dest,MPI_COMM_WORLD,ierr)
endif




! 1. split up additive coefficient vector in single coefficient vectors
! 2. multiply additive coefficient vectors with regressor matrices
! 3. do steps 1. and 2. for multiplicative coefficients (i.e. incentive1-coefficients)
! 4. convert vector of cholesky factors to cholesky-matrix
! 5. transform diagonal cholesky matrix elements to R+
call splitxbeta(COEFFS,GAMMA_RAW,LAMBDA_RAW,RHO_RAW,TAU_RAW,OMEGA_RAW,GAMMA_INC1,LAMBDA_INC1,RHO_INC1,TAU_INC1,OMEGA_INC1,CHOL_RC)




! some printing business
if (my_rank .EQ. 0) then
  ! print parameter values
  if (PRINTSWITCH(1) .AND. NEWITER .AND. CALCGRAD==.FALSE.) call printparameters
  ! print header for vector of individual likelihoods
  if (PRINTSWITCH(3)) call printlikel_intheader
endif




! MPI: partition loop 
local_N=N/proc  
if (my_rank < (proc-1)) then 
  local_ind_beg = 1 + local_N * my_rank
  local_ind_end =     local_N *(my_rank+1)
else
  local_ind_beg = 1 + local_N * my_rank
  local_ind_end = N
endif




! do more than one of the main calculations in case of gradient evaluation
do nc=1,NCALCS
   
  ! initialize local values
  LIKEL_ind=0.0d0
  local_GRADIENT_ind=0.0d0

  ! main calculations
  do ind=local_ind_beg,local_ind_end
    if (CALCGRAD==.FALSE.) then
	  ! calculate individual likelihood contribution
	  LIKEL_ind(ind)=indlogl(0.0d0)
	  ! calculate indiviual-level parameter distribution
      if (CALCIPAR) then
        call par_avg(LIKEL_ind(ind))
        call indivdistr(LIKEL_ind(ind))
      endif
    else
	  ! calculate individual-level gradients
      local_GRADIENT_ind(ind)=dfridr(indlogl_grad,0.0d0,max(abs(coeffs(nc))*1.0d-5,1.0d-6),ridd_err)
	  ridders_err=max(ridd_err,ridders_err)
    endif
  enddo




  ! merge things from different machines 
  if (CALCGRAD==.FALSE.) then

   ! start to calculate LIKEL_ind_global (only matters for root process)
   LIKEL_ind_global = LIKEL_ind

   ! start to calculate individual-level parameters (only matters for root process)
   if (CALCIPAR) then
      GAM_AVG_global=GAM_AVG
      LAM_AVG_global=LAM_AVG
      RHO_AVG_global=RHO_AVG
      TAU_AVG_global=TAU_AVG
      OMG_AVG_global=OMG_AVG
   endif




    ! Join the individual likelihoods calculated by each process.
    if (my_rank .EQ. 0 .AND. proc>1) then
      ! collect the remaining likelihood contributions, add up
      do source = 1, proc-1
        call MPI_RECV(LIKEL_ind, N, MPI_DOUBLE_PRECISION, source, tag1, MPI_COMM_WORLD, status, ierr)
        LIKEL_ind_global = LIKEL_ind_global + LIKEL_ind
      enddo
      if (CALCIPAR) then 
        do source = 1, proc-1
          call MPI_RECV(GAM_AVG, N, MPI_DOUBLE_PRECISION, source, tag4, MPI_COMM_WORLD, status, ierr)
          call MPI_RECV(LAM_AVG, N, MPI_DOUBLE_PRECISION, source, tag5, MPI_COMM_WORLD, status, ierr)
          call MPI_RECV(RHO_AVG, N, MPI_DOUBLE_PRECISION, source, tag6, MPI_COMM_WORLD, status, ierr)
          call MPI_RECV(TAU_AVG, N, MPI_DOUBLE_PRECISION, source, tag7, MPI_COMM_WORLD, status, ierr)
          call MPI_RECV(OMG_AVG, N, MPI_DOUBLE_PRECISION, source, tag8, MPI_COMM_WORLD, status, ierr)
          GAM_AVG_global=GAM_AVG_global+GAM_AVG
          LAM_AVG_global=LAM_AVG_global+LAM_AVG
          RHO_AVG_global=RHO_AVG_global+RHO_AVG
          TAU_AVG_global=TAU_AVG_global+TAU_AVG
          OMG_AVG_global=OMG_AVG_global+OMG_AVG
        enddo
      endif
 
    elseif (my_rank .NE. 0) then
      ! if not root process: send the local likelihood contributions to root process, return to main
      call   MPI_SEND(LIKEL_ind, N, MPI_DOUBLE_PRECISION, dest, tag1, MPI_COMM_WORLD, ierr)
      if (CALCIPAR) then 
        call MPI_SEND(GAM_AVG,   N, MPI_DOUBLE_PRECISION, dest, tag4, MPI_COMM_WORLD, ierr)
        call MPI_SEND(LAM_AVG,   N, MPI_DOUBLE_PRECISION, dest, tag5, MPI_COMM_WORLD, ierr)
        call MPI_SEND(RHO_AVG,   N, MPI_DOUBLE_PRECISION, dest, tag6, MPI_COMM_WORLD, ierr)
        call MPI_SEND(TAU_AVG,   N, MPI_DOUBLE_PRECISION, dest, tag7, MPI_COMM_WORLD, ierr)
        call MPI_SEND(OMG_AVG,   N, MPI_DOUBLE_PRECISION, dest, tag8, MPI_COMM_WORLD, ierr)
      endif
      return
    endif

  elseif (CALCGRAD==.TRUE.) then
    
	! Join the individual gradients calculated by each process
    if (my_rank .EQ. 0 .AND. proc>1) then
      ! add the other likelihood contributions to local_GRADIENT_ind
      do source = 1, proc-1
        call MPI_RECV( temp_GRADIENT_ind, N, MPI_DOUBLE_PRECISION, source, tag2, MPI_COMM_WORLD, status, ierr)
        local_GRADIENT_ind=local_GRADIENT_ind+temp_GRADIENT_ind
        call MPI_RECV(ridd_err,           1, MPI_DOUBLE_PRECISION, source, tag3, MPI_COMM_WORLD, status, ierr)
        ridders_err=max(ridd_err,ridders_err)
      enddo
      
      GRADIENT_ind(:,nc) = local_GRADIENT_ind
      
    elseif (my_rank .NE. 0) then
      ! if not root process: send the local likelihood contributions to root process, return to main
      call MPI_SEND(local_GRADIENT_ind, N, MPI_DOUBLE_PRECISION, dest, tag2, MPI_COMM_WORLD, ierr)
      call MPI_SEND(ridders_err,        1, MPI_DOUBLE_PRECISION, dest, tag3, MPI_COMM_WORLD, ierr)
      if (nc==NCALCS) return
    endif
  endif
enddo


! print like_int if desired
if (PRINTSWITCH(3) .AND. NEWITER) write(likel_ind_print,*) LIKEL_ind_global
      
! sum up individual likelihood contributions
LOGL_VALUE= - sum(LIKEL_ind_global)
if (LOGL_VALUE .LE. -1.0d6) LOGL_VALUE=1.0d6

! add value of log-likelihood to printout of parameters 
if (NEWITER .AND. PRINTSWITCH(1)) then
  write(parameters_print,*) LOGL_VALUE
  write(parameters_print,*)
  write(parameters_print,*)
  write(parameters_print,*)
endif


if (CALCIPAR) call printparind




! **********************************************************************

contains




! **********************************************************************

subroutine splitxbeta(COEFFS_LOCAL,GAMMA_RAW,LAMBDA_RAW,RHO_RAW,TAU_RAW,OMEGA_RAW,GAMMA_INC1,LAMBDA_INC1,RHO_INC1,TAU_INC1,OMEGA_INC1,CHOL_RC)
! 1. split up additive coefficient vector in single coefficient vectors
! 2. multiply additive coefficient vectors with regressor matrices
! 3. do steps 1. and 2. for multiplicative coefficients (i.e. incentive1-coefficients)
! 4. convert vector of cholesky factors to cholesky-matrix
! 5. transform diagonal cholesky matrix elements to R+
real(DP),intent(in)::COEFFS_LOCAL(K)
real(DP),intent(out)::GAMMA_RAW(NJ), LAMBDA_RAW(NJ), RHO_RAW(NJ), TAU_RAW(NJ), OMEGA_RAW(NJ),&
                      GAMMA_INC1(NJ),LAMBDA_INC1(NJ),RHO_INC1(NJ),TAU_INC1(NJ),OMEGA_INC1(NJ),&
                      CHOL_RC(L_COV_RC,L_COV_RC)
integer(I4B)::counter1,counter2

! 1. split up additive coefficient vector in single coefficient vectors
! 2. multiply additive coefficient vectors with regressor matrices
counter1=0
if (K_GAMMA >0)		then; GAMMA_RAW =matmul(X_GAMMA ,COEFFS_LOCAL(counter1+1:counter1+K_GAMMA )); counter1=counter1+K_GAMMA 
else; GAMMA_RAW =0.0d0;	endif
if (K_LAMBDA>0)		then; LAMBDA_RAW=matmul(X_LAMBDA,COEFFS_LOCAL(counter1+1:counter1+K_LAMBDA)); counter1=counter1+K_LAMBDA
else; LAMBDA_RAW=1.0d0;	endif;
if (K_RHO   >0)		then; RHO_RAW   =matmul(X_RHO   ,COEFFS_LOCAL(counter1+1:counter1+K_RHO   )); counter1=counter1+K_RHO
else; RHO_RAW   =1.0d0;	endif
if (K_TAU   >0)		then; TAU_RAW   =matmul(X_TAU   ,COEFFS_LOCAL(counter1+1:counter1+K_TAU   )); counter1=counter1+K_TAU
else; TAU_RAW   =1.0d0;	endif
if (K_OMEGA >0)		then; OMEGA_RAW =matmul(X_OMEGA ,COEFFS_LOCAL(counter1+1:counter1+K_OMEGA )); counter1=counter1+K_OMEGA 
else; OMEGA_RAW =0.0d0;	endif

! 3. do steps 1. and 2. for multiplicative coefficients (i.e. incentive1-coefficients)
if (K_GAMMA >0 .AND. K_INC1>0) then;  GAMMA_INC1 =matmul(X_INC1 ,COEFFS_LOCAL(counter1+1:counter1+1)); counter1=counter1+1
else;  GAMMA_INC1 =1.0d0;	endif
if (K_LAMBDA>0 .AND. K_INC1>0) then; LAMBDA_INC1 =matmul(X_INC1 ,COEFFS_LOCAL(counter1+1:counter1+1)); counter1=counter1+1
else; LAMBDA_INC1 =1.0d0;	endif
if (K_RHO   >0 .AND. K_INC1>0) then;    RHO_INC1 =matmul(X_INC1 ,COEFFS_LOCAL(counter1+1:counter1+1)); counter1=counter1+1
else;    RHO_INC1 =1.0d0;	endif
if (K_TAU   >0 .AND. K_INC1>0) then;    TAU_INC1 =matmul(X_INC1 ,COEFFS_LOCAL(counter1+1:counter1+1)); counter1=counter1+1
else;    TAU_INC1 =1.0d0;	endif
if (K_OMEGA >0 .AND. K_INC1>0) then;  OMEGA_INC1 =matmul(X_INC1 ,COEFFS_LOCAL(counter1+1:counter1+1)); counter1=counter1+1
else;  OMEGA_INC1 =1.0d0;	endif

! 4. convert vector of cholesky factors to cholesky-matrix
! 5. transform diagonal cholesky matrix elements to R+
if (K_COV_RC>0) then 
  ! convert cholesky vector to matrix
  if (COVREST) then
    do counter2=1,K_COV_RC
      CHOL_RC(counter2,counter2)=COEFFS_LOCAL(counter1+counter2)
    enddo
  else
    call vec2triangmat(L_COV_RC,COEFFS_LOCAL(counter1+1:counter1+K_COV_RC),CHOL_RC) 
  endif
  ! transform diagonal elements to R+
  do counter2=1,L_COV_RC
    CHOL_RC(counter2,counter2)=exp(CHOL_RC(counter2,counter2))
  enddo
endif



end subroutine splitxbeta




!****************************************************************************

function indlogl_grad(x)
! computes the individual log-likelihood contribution for dfridr
real(DP),intent(in)::x
real(DP)::indlogl_grad
real(DP)::COEFFS_GRAD(K)


COEFFS_GRAD=COEFFS
COEFFS_GRAD(nc)=COEFFS_GRAD(nc)+x

call splitxbeta(COEFFS_GRAD,GAMMA_RAW,LAMBDA_RAW,RHO_RAW,TAU_RAW,OMEGA_RAW,GAMMA_INC1,LAMBDA_INC1,RHO_INC1,TAU_INC1,OMEGA_INC1,CHOL_RC)

indlogl_grad=indlogl(0.0d0)

end function indlogl_grad




!****************************************************************************

function indlogl(x)
! computes the individual log-likelihood contribution
! only implemented like this to ensure compatibility with indlogl_grad
real(DP),intent(in)::x
real(DP)::indlogl
real(DP)::integrand_vec(NDRAWS)
real(DP)::dummy(L_COV_RC)
real(DP),dimension(INDIVINDIC(ind,4)):: GAM,LAM,RHO,TAU,OMG

if (L_COV_RC==0) then
  
  ! transform parameters to final values
  call finalpars(dummy,GAM,LAM,RHO,TAU,OMG)
  ! calculate the individual integrand
  integrand_vec(1)=integrand(GAM,LAM,RHO,TAU,OMG)
  ! calculate the individual log-likelihood
  indlogl=log(integrand_vec(1))

else

  ! transform (0,1) draws to N(0,Sigma)
  nodes=matmul(transpose(CHOL_RC),SIM(ind,:,:))
  ! calculate likelihood at each draw
  do r=1,NDRAWS
    ! add unobserved heterogeneity to parameters, transform to final value
    call finalpars(nodes(:,r),GAM,LAM,RHO,TAU,OMG)
    ! calculate the individual integrand at each simulation draw
    integrand_vec(r)=integrand(GAM,LAM,RHO,TAU,OMG)
  enddo
  ! sum over all integration points and take the logarithm
  ! to obtain individual log-likelihood
  indlogl=log(sum(integrand_vec)/(NDRAWS+.0d0))
endif

end function indlogl




!****************************************************************************

subroutine finalpars(HET_PARS,GAM,LAM,RHO,TAU,OMG)
! adds heterogeneous part to parameters and converts them to final form

use nrtype
use controlpars
implicit none
real(DP),intent(in):: HET_PARS(L_COV_RC)
real(DP),intent(out),dimension(INDIVINDIC(ind,4)):: GAM,LAM,RHO,TAU,OMG
integer(I4B)::j


! add heterogeneous part to parameters if necessary
j=1
if (HET_GAMMA) then
  GAM= GAMMA_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))+HET_PARS(j); 	j=j+1
else 
  GAM= GAMMA_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))
endif

if (HET_LAMBDA) then
  LAM=LAMBDA_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))+HET_PARS(j);	j=j+1
else
  LAM=LAMBDA_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))
endif

if (HET_RHO) then
  RHO=   RHO_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))+HET_PARS(j);	j=j+1
else
  RHO=   RHO_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))
endif

if (HET_TAU) then
  TAU=   TAU_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))+HET_PARS(j);	j=j+1
else 
  TAU=   TAU_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))
endif

if (HET_OMEGA) then
  OMG= OMEGA_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))+HET_PARS(j);	j=j+1
else
  OMG= OMEGA_RAW(INDIVINDIC(ind,2):INDIVINDIC(ind,3))
endif




! multiply with incentive_1 - coefficient
if (K_INC1>0 .AND. X_INC1(INDIVINDIC(ind,2),1)>1.0e-5) then
  if (K_GAMMA >0) GAM=GAM*exp( GAMMA_INC1(INDIVINDIC(ind,2):INDIVINDIC(ind,3)))
  if (K_LAMBDA>0) LAM=LAM*exp(LAMBDA_INC1(INDIVINDIC(ind,2):INDIVINDIC(ind,3)))
  if (K_RHO   >0) RHO=RHO*exp(   RHO_INC1(INDIVINDIC(ind,2):INDIVINDIC(ind,3)))
  if (K_TAU   >0) TAU=TAU*exp(   TAU_INC1(INDIVINDIC(ind,2):INDIVINDIC(ind,3)))
  if (K_OMEGA >0) OMG=OMG*exp( OMEGA_INC1(INDIVINDIC(ind,2):INDIVINDIC(ind,3)))
endif




! convert parameters from R^1 to restricted form
if (K_GAMMA >0) call convertcoeffs(INDIVINDIC(ind,4),EXP_GAMMA ,CDF_GAMMA ,CDF2_GAMMA,GAM)
if (K_LAMBDA>0) call convertcoeffs(INDIVINDIC(ind,4),EXP_LAMBDA,CDF_LAMBDA,.FALSE.   ,LAM)
if (K_RHO   >0) call convertcoeffs(INDIVINDIC(ind,4),EXP_RHO   ,CDF_RHO   ,.FALSE.   ,RHO)
if (K_TAU   >0) call convertcoeffs(INDIVINDIC(ind,4),EXP_TAU   ,CDF_TAU   ,.FALSE.   ,TAU)
if (K_OMEGA >0) call convertcoeffs(INDIVINDIC(ind,4),EXP_OMEGA ,CDF_OMEGA ,.FALSE.   ,OMG)

end subroutine finalpars




!****************************************************************************

function integrand(GAM,LAM,RHO,TAU,OMG)
! computes the individual integrand	

use nrtype
use controlpars
use hm_utils
use nrutil
implicit none
real(DP):: integrand
real(DP),dimension(INDIVINDIC(ind,4)):: GAM,LAM,RHO,TAU,OMG,u,integrand_j
real(DP),external::cediff_cara,cediff_crra
integer(I4B)::j,indj

! calculate `integrand_j', i.e. the value of the integrand at each choice j=1,...,J_i
do j=1,indivindic(ind,4)

  ! set counter for the data matrices
  indj=INDIVINDIC(ind,2)+j-1

  ! calculate inner part of the logit function
  if (CARA==.TRUE.) then 
    u(j)=(2.0d0*Y(indj,1)-1)*(1.0d0/TAU(j))*cediff_cara(GAM(j),LAM(j),RHO(j),X_GENERAL(indj,:))
  else
    u(j)=(2.0d0*Y(indj,1)-1)*(1.0d0/TAU(j))*cediff_crra(GAM(j),LAM(j),RHO(j),X_GENERAL(indj,:))
  endif

  ! make sure it is finite, otherwise: warning
  if (isnan(u(j)) .AND. CALCIPAR==.FALSE.) then
    write(*,*) 'NaN for integrand of indiv/choice#/draw/function_eval:',ind,j,r,nf
    u(j)=tiny(integrand_j)
  endif
  ! take account of tremble parameter and apply logistic transform to u(.) for the integrand_j
  integrand_j(j) = (1-OMG(j)) * logistcdf(u(j)) + OMG(j)/2.0d0

enddo 

! make sure individual likelihood contribution is >0, otherwise: warning
if (product(integrand_j)<tiny(integrand_j)) write(*,*) 'product(integrand_j)<tiny(integrand_j) ind, nf:',ind,nf
integrand=max( product(integrand_j), tiny(integrand_j) ) 

end function integrand




! **********************************************************************

subroutine printparameters
! prints the current parameter values
real(DP)::COEFFS_TEMP(K),CHOL_TEMP(L_COV_RC,L_COV_RC),CORR_TEMP(L_COV_RC,L_COV_RC)
integer(I4B)::i

COEFFS_TEMP=COEFFS

! print out current function evaluation
write(parameters_print,*) nf

! transform restricted coefficients to their true values
call convertcoeffs_end(COEFFS_TEMP)

! transform cholesky-factors to correlations/standard deviations
if (COVREST==.FALSE.) then
  call vec2triangmat(L_COV_RC,COEFFS_TEMP(K-K_COV_RC+1:K),CHOL_TEMP)
  call chol2corr(L_COV_RC,CHOL_TEMP,CORR_TEMP)
  call triangmat2vec(L_COV_RC,CORR_TEMP,COEFFS_TEMP(K-K_COV_RC+1:K))
endif

! print the transformed parameters
do i=1,K
  write(parameters_print,*) COEFFS_TEMP(i)
enddo
write(parameters_print,*) ' '

end subroutine printparameters





! **********************************************************************

subroutine printlikel_intheader
! prints the header for individual likelihood contributions

write(likel_ind_print,*)
write(likel_ind_print,*)
write(likel_ind_print,*)
write(likel_ind_print,*) 'Individual   Likelihood_Contribution   Function_Count',funccounter
write(likel_ind_print,*)

end subroutine printlikel_intheader




! **********************************************************************

subroutine printparind
! prints individual-level parameter averages and distributions
integer(I4B)::j


if (HET_GAMMA) then
  open (UNIT=1002,FILE=paroutpathind//'gam_avg.csv')
  do j=1,N
    write(1002,*) indivindic(j,1),';',GAM_AVG_global(j)
  enddo
endif
if (HET_LAMBDA) then
  open (UNIT=1002,FILE=paroutpathind//'lam_avg.csv')
  do j=1,N
    write(1002,*) indivindic(j,1),';',LAM_AVG_global(j)
  enddo
endif
if (HET_RHO) then
  open (UNIT=1002,FILE=paroutpathind//'rho_avg.csv')
  do j=1,N
    write(1002,*) indivindic(j,1),';',RHO_AVG_global(j)
  enddo
endif
if (HET_TAU) then
  open (UNIT=1002,FILE=paroutpathind//'tau_avg.csv')
  do j=1,N
    write(1002,*) indivindic(j,1),';',TAU_AVG_global(j)
  enddo
endif
if (HET_OMEGA) then
  open (UNIT=1002,FILE=paroutpathind//'omg_avg.csv')
  do j=1,N
    write(1002,*) indivindic(j,1),';',OMG_AVG_global(j)
  enddo
endif

end subroutine printparind




!****************************************************************************

subroutine par_avg(loglikelind)
! computes averages of individual-level parameters
implicit none
real(DP),intent(in)::loglikelind
real(DP),dimension(NDRAWS)::integrand_vec,GAM_TEMP,LAM_TEMP,RHO_TEMP,TAU_TEMP,OMG_TEMP
real(DP),dimension(INDIVINDIC(ind,4)):: GAM,LAM,RHO,TAU,OMG
real(DP)::COV_RC(L_COV_RC,L_COV_RC)
integer(I4B)::j,r

! do not calculate anything if no individual-level parameters present
if (L_COV_RC==0) return 

! transform (0,1) draws to N(0,Sigma)
nodes=matmul(transpose(CHOL_RC),SIM(ind,:,:))
! calculate likelihood at each draw
do r=1,NDRAWS
  ! add unobserved heterogeneity to parameters, transform to final value
  call finalpars(nodes(:,r),GAM,LAM,RHO,TAU,OMG)
  ! calculate the individual integrand at each simulation draw
  integrand_vec(r)=integrand(GAM,LAM,RHO,TAU,OMG)
  ! multiply with respective parameter to get individual average
  if (HET_GAMMA)  GAM_TEMP(r)=GAM(1)*integrand_vec(r)
  if (HET_LAMBDA) then
    if (maxval(LAM)>1.000000001d0) then
      LAM_TEMP(r)=maxval(LAM)*integrand_vec(r)
    else 
      LAM_TEMP(r)=minval(LAM)*integrand_vec(r)
    endif
  endif
  if (HET_RHO)    RHO_TEMP(r)=RHO(1)*integrand_vec(r)
  if (HET_TAU)    TAU_TEMP(r)=TAU(1)*integrand_vec(r)
  if (HET_OMEGA)  OMG_TEMP(r)=OMG(1)*integrand_vec(r)
enddo

if (HET_GAMMA)  GAM_AVG(ind)=sum(GAM_TEMP)/(NDRAWS+.0d0)/exp(loglikelind)
if (HET_LAMBDA) LAM_AVG(ind)=sum(LAM_TEMP)/(NDRAWS+.0d0)/exp(loglikelind)
if (HET_RHO)    RHO_AVG(ind)=sum(RHO_TEMP)/(NDRAWS+.0d0)/exp(loglikelind)
if (HET_TAU)    TAU_AVG(ind)=sum(TAU_TEMP)/(NDRAWS+.0d0)/exp(loglikelind)
if (HET_OMEGA)  OMG_AVG(ind)=sum(OMG_TEMP)/(NDRAWS+.0d0)/exp(loglikelind)

end subroutine par_avg




! **********************************************************************

subroutine indivdistr(loglikelind)
! computes and prints distributions of individual-level parameters
implicit none
real(DP),intent(in)::loglikelind
real(DP)::integrand_vec,GAM_TEMP,LAM_TEMP,RHO_TEMP,TAU_TEMP,OMG_TEMP
real(DP),dimension(INDIVINDIC(ind,4)):: GAM,LAM,RHO,TAU,OMG
real(DP)::COV_RC(L_COV_RC,L_COV_RC),muL(L_COV_RC)
integer(I4B)::j,LOC_GAM,LOC_LAM,LOC_RHO,LOC_TAU,LOC_OMG
real(DP),allocatable::nodesA(:,:),nodesB(:,:),covA(:,:),covB(:,:),covAB(:,:),muA(:),muB(:)
real(DP)::GAM_f(NDRAWS,2),LAM_f(NDRAWS,2),RHO_f(NDRAWS,2),TAU_f(NDRAWS,2),OMG_f(NDRAWS,2)

! do not calculate anything if no individual-level parameters present
if (L_COV_RC==0) return 

! pick individuals for whom individual-level distributionss are to be calculated
do j=1,size(indlevnr)
  if (indlevnr(j)==indivindic(ind,1)) exit
  if (j==size(indlevnr)) return
enddo
! only end up at this point if this is one of the individuals in question



! get means of parameters
muL=0.0d0
call finalpars(muL,GAM,LAM,RHO,TAU,OMG)


! get the unpartitioned vcov-matrix and locations of parameters
COV_RC=matmul(transpose(CHOL_RC),CHOL_RC)
j=1
if (HET_GAMMA)  then
  LOC_GAM=j
  muL(j)=GAM(1)
  j=j+1
endif
if (HET_LAMBDA) then
  LOC_LAM=j
  if (maxval(LAM)>1.0d0) then
    muL(j)=maxval(LAM)
  else
    muL(j)=minval(LAM)
  endif
  j=j+1
endif
if (HET_RHO)    then
  LOC_RHO=j
  muL(j)=RHO(1)
  j=j+1
endif
if (HET_TAU)    then
  LOC_TAU=j
  muL(j)=TAU(1)
  j=j+1
endif
if (HET_OMEGA)  then
  LOC_OMG=j
  muL(j)=OMG(1)
endif




! allocate partitioned mean & vcov-matrices for unidimensional conditional distributions
if (allocated(covA))  deallocate(covA); allocate(covA(1:1,1:1))
if (allocated(covB))  deallocate(covB); allocate(covB(1:L_COV_RC-1,1:L_COV_RC-1))
if (allocated(covAB)) deallocate(covAB);allocate(covAB(1:1,1:L_COV_RC-1))
if (allocated(muA))   deallocate(muA);  allocate(  muA(1:1))
if (allocated(muB))   deallocate(muB);  allocate(  muB(1:L_COV_RC-1))

! estimate unidimensional conditional distributions
if (HET_GAMMA)  then
  muA=muL(LOC_GAM)
  muB(1:LOC_GAM-1)=muL(1:LOC_GAM-1)
  muB(LOC_GAM:L_COV_RC-1)=muL(LOC_GAM+1:L_COV_RC)
  call partition_cov(1,L_COV_RC,LOC_GAM,0,COV_RC,covA,covAB,covB)
  call     conddistr(1,L_COV_RC,LOC_GAM,0,       covA,covAB,covB,muA,muB,loglikelind,'GAM')
endif
if (HET_LAMBDA) then
  muA=muL(LOC_LAM)
  muB(1:LOC_LAM-1)=muL(1:LOC_LAM-1)
  muB(LOC_LAM:L_COV_RC-1)=muL(LOC_LAM+1:L_COV_RC)
  call partition_cov(1,L_COV_RC,LOC_LAM,0,COV_RC,covA,covAB,covB)
  call     conddistr(1,L_COV_RC,LOC_LAM,0,       covA,covAB,covB,muA,muB,loglikelind,'LAM')
endif
if (HET_RHO) then
  muA=muL(LOC_RHO)
  muB(1:LOC_RHO-1)=muL(1:LOC_RHO-1)
  muB(LOC_RHO:L_COV_RC-1)=muL(LOC_RHO+1:L_COV_RC)
  call partition_cov(1,L_COV_RC,LOC_RHO,0,COV_RC,covA,covAB,covB)
  call     conddistr(1,L_COV_RC,LOC_RHO,0,       covA,covAB,covB,muA,muB,loglikelind,'RHO')
endif
if (HET_TAU) then
  muA=muL(LOC_TAU)
  muB(1:LOC_TAU-1)=muL(1:LOC_TAU-1)
  muB(LOC_TAU:L_COV_RC-1)=muL(LOC_TAU+1:L_COV_RC)
  call partition_cov(1,L_COV_RC,LOC_TAU,0,COV_RC,covA,covAB,covB)
  call     conddistr(1,L_COV_RC,LOC_TAU,0,       covA,covAB,covB,muA,muB,loglikelind,'TAU')
endif
if (HET_OMEGA) then
  muA=muL(LOC_OMG)
  muB(1:LOC_OMG-1)=muL(1:LOC_OMG-1)
  muB(LOC_OMG:L_COV_RC-1)=muL(LOC_OMG+1:L_COV_RC)
  call partition_cov(1,L_COV_RC,LOC_OMG,0,COV_RC,covA,covAB,covB)
  call     conddistr(1,L_COV_RC,LOC_OMG,0,       covA,covAB,covB,muA,muB,loglikelind,'OMG')
endif




end subroutine indivdistr




! **********************************************************************

subroutine conddistr(DIM,L,LOC1,LOC2,covA,covAB,covB,muA,muB,loglikelind,PARNAME)
! calculates the distribution of the parameters in A 
! conditional on those in B
! pdf-part not implemented for 2 parameters !!!
use controlpars
implicit none
integer(I4B),intent(in)::DIM,L,LOC1,LOC2
real(DP),intent(in)::covA(DIM,DIM),covAB(DIM,L-DIM),              &
                     covB(L-DIM,L-DIM),loglikelind,               &
                     muA(DIM),muB(L-DIM)
character(len=*)  PARNAME
character(len=3)  PARNAME_LOWER
character(len=20) indnumber
real(DP)::PAR_f(DIM+1,NDRAWS)
integer(I4B)::rA,rB,j,i1,i2
logical(LGT)::pdflag, EXP_PAR, CDF_PAR, CDF2_PAR

real(DP)::cholA(DIM,DIM),cholAB(DIM,L-DIM),cholB(L-DIM,L-DIM),    &
          nodesAraw(DIM,NDRAWS),nodesBraw(L-DIM,NDRAWS),          &
          nodesA(DIM,NDRAWS),nodesB(L-DIM,1),                     &
		  muBcondA(L-DIM,1),covBcondA(L-DIM,L-DIM),               &
		  cholBcondA(L-DIM,L-DIM),HET_PARS(L),                    &
          condintegrand_vec(NDRAWS),zeros(DIM),pdfA

real(DP),dimension(INDIVINDIC(ind,4)):: GAM,LAM,RHO,TAU,OMG


! ground transformation parameters
EXP_PAR =.FALSE.
CDF_PAR =.FALSE.
CDF2_PAR=.FALSE.

! first get the partitioned nodes
i1=1;i2=1
do j=1,L
  if (j==LOC1 .OR. j==LOC2) then
    nodesAraw(i1,:)=SIM(ind,j,:)
	i1=i1+1
  else
    nodesBraw(i2,:)=SIM(ind,j,:)
	i2=i2+1
  endif
enddo




! convert partition A draws to draws from normal
call choldc(DIM,covA,cholA,pdflag)
nodesA=matmul(transpose(cholA),nodesAraw)
! generate matrix of cholesky-factors for conditional covariance matrix B
covBcondA=covB - matmul(matmul(transpose(covAB),covA),covAB)
call choldc(L-DIM,covBcondA,cholBcondA,pdflag)

! main loop: 
do rA=1,NDRAWS
  
  ! calculate conditional average of nodes to be integrated out
  muBcondA=matmul(matmul(transpose(covAB),covA),nodesA(:,rA:rA))

  ! integrate out B-parameters for each draw of A
  do rB=1,NDRAWS

    ! calculate nodes for parts that are to be integrated out
    nodesB=muBcondA+matmul(transpose(cholBcondA),nodesBraw(:,rB:rB))
    
	! get nodes in the right order
    i1=1;i2=1
    do j=1,L
      if (j==LOC1 .OR. j==LOC2) then
        HET_PARS(j)=nodesA(i1,rA)
        i1=i1+1
      else
        HET_PARS(j)=nodesB(i2,1)
        i2=i2+1
      endif
    enddo

    ! transform parameters to final values
    call finalpars(HET_PARS,GAM,LAM,RHO,TAU,OMG)
    ! calculate the individual integrand at each simulation draw
    condintegrand_vec(rB)=integrand(GAM,LAM,RHO,TAU,OMG)
  
  enddo

  ! fill first columns of PAR_f with values of A-parameters
  if     (PARNAME=='GAM')  then
    PARNAME_LOWER = 'gam'
    PAR_f(1,rA)=GAM(1)
	if (EXP_GAMMA)  EXP_PAR =.TRUE.
	if (CDF_GAMMA)  CDF_PAR =.TRUE.
	if (CDF2_GAMMA) CDF2_PAR=.TRUE.
  elseif (PARNAME=='LAM') then
    PARNAME_LOWER = 'lam'
    if (maxval(LAM)>1.0d0) then
      PAR_f(1,rA)=maxval(LAM)
	else
	  PAR_f(1,rA)=minval(LAM)
	endif
	if (EXP_LAMBDA) EXP_PAR =.TRUE.
	if (CDF_LAMBDA) CDF_PAR =.TRUE.
  elseif (PARNAME=='RHO')    then
    PARNAME_LOWER = 'rho'
    PAR_f(1,rA)=RHO(1)
	if (EXP_RHO)    EXP_PAR =.TRUE.
	if (CDF_RHO)    CDF_PAR =.TRUE.
  elseif (PARNAME=='TAU')    then
    PARNAME_LOWER = 'tau'
    PAR_f(1,rA)=TAU(1)
	if (EXP_TAU)    EXP_PAR =.TRUE.
	if (CDF_TAU)    CDF_PAR =.TRUE.
  elseif (PARNAME=='OMG')  then
    PARNAME_LOWER = 'omg'
    PAR_f(1,rA)=OMG(1)
	if (EXP_OMEGA)  EXP_PAR =.TRUE.
	if (CDF_OMEGA)  CDF_PAR =.TRUE.
  else
    PAR_f=0.0d0
  endif




  ! calculate pdf value of A
  if     (EXP_PAR)  then
    pdfA=   lognpdf(exp(nodesA(1,rA)+log(muA(1))),log(muA(1)),cholA(1,1))
  elseif (CDF_PAR)  then
    pdfA= logitnpdf(logistcdf(nodesA(1,rA)+log(muA(1)/(1-muA(1)))),log(muA(1)/(1-muA(1))),cholA(1,1))
  elseif (CDF2_PAR) then
    pdfA=logit2npdf((logistcdf(nodesA(1,rA)+log((1+muA(1))/(1-muA(1))))-.5d0)*2.0d0,log((1+muA(1))/(1-muA(1))),cholA(1,1))
  else
    pdfA=npdf(DIM,nodesA(:,rA),cholA)
  endif

  ! fill last column of PAR_f with the conditional distribution
  PAR_f(DIM+1,rA)=pdfA*sum(condintegrand_vec)/(NDRAWS+.0d0)/exp(loglikelind)

enddo


if      (indivindic(ind,1)-       10<0) then
  write(indnumber,'(I1)') indivindic(ind,1)
elseif  (indivindic(ind,1)-      100<0) then
  write(indnumber,'(I2)') indivindic(ind,1)
elseif  (indivindic(ind,1)-     1000<0) then
  write(indnumber,'(I3)') indivindic(ind,1)
elseif  (indivindic(ind,1)-    10000<0) then
  write(indnumber,'(I4)') indivindic(ind,1)
elseif  (indivindic(ind,1)-   100000<0) then
  write(indnumber,'(I5)') indivindic(ind,1)
elseif  (indivindic(ind,1)-  1000000<0) then
  write(indnumber,'(I6)') indivindic(ind,1)
elseif  (indivindic(ind,1)- 10000000<0) then
  write(indnumber,'(I7)') indivindic(ind,1)
elseif  (indivindic(ind,1)-100000000<0) then
  write(indnumber,'(I8)') indivindic(ind,1)
else
  write(indnumber,'(I9)') indivindic(ind,1)
endif

open (UNIT=1003,FILE=paroutpathind//trim(indnumber)//'_'//PARNAME_LOWER//'_f'//'.csv')
call printarraycsv(NDRAWS,DIM+1,transpose(PAR_f),1003)
close(1003)


end subroutine conddistr




! **********************************************************************

end subroutine logl_all
