!  rcmain.f90 
!
!
!  FUNCTIONS:
!
!	rcmain      - Entry point of console application.
!

!****************************************************************************
!
!  PROGRAM: rcmain
!
!  PURPOSE: estimates rc model for preferences in CentERpanel
!
!****************************************************************************


program rcmain
use controlpars
use hm_utils
implicit none

real(DP)::COEFFS(K)	! coefficient vector 
integer(I4B), parameter :: LIV = 60             ! optimisation control parameter for overall log-likelihood
integer(I4B), parameter :: LV  = 77+K*(K+17)/2  ! optimisation control parameter for overall log-likelihood
integer(I4B):: IV(LIV)                          ! optimisation control parameter for overall log-likelihood
real(DP)::     V(LV)                            ! optimisation control parameter for overall log-likelihood
integer(I4B):: uip(1),i,j               
real(DP):: urp(1),SE(K,K),dummy
external logl_all



! Let the system do what it needs to start up MPI.
call MPI_INIT(ierr)
! Get my process rank.
call MPI_COMM_RANK(MPI_COMM_WORLD, my_rank, ierr)
! Find out how many processes are being used.
call MPI_COMM_SIZE(MPI_COMM_WORLD, proc, ierr)



! load the data
call loaddata
if (L_COV_RC>0) call gethalton

! check whether model is specified consistently, convert start values
call checkcontrols




if (my_rank .EQ. 0) then

  ! control parameter for the loop of processes of rank>0
  paralliter=.TRUE.
  ! do not calculate gradient yet
  CALCGRAD=.FALSE.
  ! do not calculate individual-specific parameters yet
  CALCIPAR=.FALSE.

  ! set optimisation control parameters
  call optimcontrol
  COEFFS=COEFFS_START


  ! control printing
  if (PRINTSWITCH(1)) open (UNIT=parameters_print,FILE=paroutpath//'par.txt')
  if (PRINTSWITCH(2)) open (UNIT=iv(21),          FILE=paroutpath//'max.txt')
  if (PRINTSWITCH(3)) open (UNIT=likel_ind_print, FILE=paroutpath//'likel_ind.txt')


  
  ! maximise likelihood: first round
  if (MAX1) then
    funccounter=0
    call smsno(K, COEFFS_SCALE, COEFFS, LOGL_ALL, iv, liv, lv, v, uip, urp, LOGL_ALL)
  endif

  ! maximise likelihood: second round
  if (MAX2) then
    ! reset optimisation control parameters
    call optimcontrol
    do i=1,K
      COEFFS_SCALE(i)=1.0d0/max(abs(COEFFS(i)),2.0d-3)
    enddo
	
	funccounter=0
    call smsno(K, COEFFS_SCALE, COEFFS, LOGL_ALL, iv, liv, lv, v, uip, urp, LOGL_ALL)
  endif


  ! close some of the files used for output
  if (PRINTSWITCH(3)) close(likel_ind_print)
  if (PRINTSWITCH(2)) close(iv(21))



  ! calculate gradient
  if (STDE) then
    CALCGRAD=.TRUE.
    call logl_all(K,COEFFS,0,dummy,uip,urp,urp)
    call standerr(COEFFS,SE,v(10))
  endif

  ! calculate individual-specific parameters
  if (IPAR) then
	CALCGRAD=.FALSE.
    CALCIPAR=.TRUE.
    call logl_all(K,COEFFS,0,dummy,uip,urp,urp)
  endif


  ! close last output file
  if (PRINTSWITCH(1)) close(1)

  ! stop the remaining processes
  paralliter=.FALSE.
  call MPI_BCAST(PARALLITER,  1,MPI_LOGICAL,         0,MPI_COMM_WORLD,ierr)

else

  ! control parameter for the loop of processes of rank>0
  PARALLITER=.TRUE.

  do while (PARALLITER)
    ! receive control parameter for this loop
    call MPI_BCAST(PARALLITER,1,MPI_LOGICAL,         0,MPI_COMM_WORLD,ierr)
	if (PARALLITER .EQ. .FALSE.) exit
    ! receive control parameter for gradients / individual-level parameters
    call MPI_BCAST(CALCGRAD  ,1,MPI_LOGICAL,         0,MPI_COMM_WORLD,ierr)
    call MPI_BCAST(CALCIPAR  ,1,MPI_LOGICAL,         0,MPI_COMM_WORLD,ierr)
    ! receive current coefficient matrix
    call MPI_BCAST(    COEFFS,K,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
    ! calculate partition of the likelihood contribution
    call logl_all(K,COEFFS,0,v(10),uip,urp,urp)
  enddo

endif

call MPI_FINALIZE(ierr)



! **********************************************************************

contains 



! **********************************************************************

subroutine optimcontrol
! controls optimisation parameters for maximisation of log-likelihood
external:: deflt

call deflt(2, iv, liv, lv, v)

if (PRINTSWITCH(2)) then
  iv(21)=11
else
  iv(21)=0
endif

! change defaults
iv(17)=1500                 ! maximum number of function evaluations
iv(18)=1200                 ! maximum number of iterations
v(32)=max(1.0d-7,v(32))		! relative function convergence tolerance 
v(35)=.1d-1					! maximum 2-norm for the first scaled step length

end subroutine optimcontrol




! **********************************************************************

subroutine gethalton
! controls optimisation parameters for maximisation of log-likelihood
integer(I4B),parameter::HALLENGTH=NDRAWS*N
integer(I4B)::STEP,SEED(L_COV_RC),LEAP(L_COV_RC),BASE(L_COV_RC),i1,i2,i3,i4
real(DP)::DRAWVEC(L_COV_RC,HALLENGTH)


! the first `SEED' draws are discarded from each sequence

STEP=0

SEED(1)=99
LEAP(1)=1
BASE(1)=2
if (L_COV_RC>1) then
  SEED(2)=99
  LEAP(2)=1
  BASE(2)=3
  if (L_COV_RC>2) then
    SEED(3)=99
    LEAP(3)=1
    BASE(3)=5
    if (L_COV_RC>3) then
      SEED(4)=99
      LEAP(4)=1
      BASE(4)=7
      if (L_COV_RC>4) then
        SEED(5)=99
        LEAP(5)=1
        BASE(5)=11
	   endif
     endif
   endif
endif



call i4_to_halton_sequence (L_COV_RC, HALLENGTH, STEP, SEED, LEAP, BASE, DRAWVEC)

do i1=1,L_COV_RC
  do i2=1,HALLENGTH
    DRAWVEC(i1,i2)=ncdfinv(DRAWVEC(i1,i2))
  enddo
enddo


do i1=1,L_COV_RC
  i2=1
  i3=1
  do i2=1,N
    do i3=1,NDRAWS
	  SIM(i2,i1,i3)=DRAWVEC(i1,(i2-1)*NDRAWS+i3)
	enddo
  enddo
enddo



end subroutine gethalton




! **********************************************************************

end program rcmain
