from SCons.Script import *
from glob import glob
import re

def generate(env):
    '''
    Add a very simple builder for fmopt. Requires env['FMOPT_MODE']  to be 
    set to either 'SERIAL' or 'MPI'
    
    '''

    # Define builder 
    fmoptbuilder = Builder(action =
                           Action(compilefmopt,
                                  "\n\nbuilding '$TARGET' from '$SOURCE'",
                                  varlist = ['FMOPT_MODE', 'DEBUG']),
                           chdir = True,
                           emitter = modify_targets_sources)
    # Add builder to environment
    env.Append(BUILDERS = {'fmopt' : fmoptbuilder})

    return


def modify_targets_sources(target, source, env):
    """
    Provides list of intermediate targets (*.o and *.gpi files) and dependencies 
    for fmopt compilation. The added dependencies consist of: 
      -- The fmopt_shared.global and fmopt_user.global files.
      -- The target_body.start and target_body.options files. 
      -- The contents of the fmopt2 directory.
    The body of the target file name must be the same as the main source file
    name, and the .start and .options files. I.e. we need: 
    
    [foo.c, foo.start, foo.options]        to produce        foo[.out]
    
    """

    # Get the body of the target name, neglecting any file extension if present
    target_path, target_name = os.path.split(target[0].abspath)
    target_path = target_path + os.path.sep
    target_body = os.path.splitext(target_name)[0]

    # would like to add all source files with .c extension with .o extension, 
    # but cannot do so for those already built for tests. so take care of that
    # in compilefmopt below.
    target.append(target_path + target_body + '.o')

    # Intermediate targets that are the same for MPI / SERIAL compilation
    target.extend([target_path + 'fmopt.o',
                   target_path + 'fmopt_cdfn.o',
                   target_path + 'fmopt_client.o',
                   target_path + 'fmopt_from_c.o',
                   target_path + 'fmopt_mat.o',
                   target_path + 'fmopt_server.o',
                   target_path + 'gpc.o',
                   target_path + 'gpc_bp.gpi',
                   target_path + 'gpc_cp.gpi',
                   target_path + 'gpc_delphi.gpi',
                   target_path + 'gpc_ep.gpi',
                   target_path + 'gpc-all.gpi',
                   target_path + 'gpc.gpi'])

    # Append relevant target depending on MPI / SERIAL compilation  
    if env['FMOPT_MODE'] == 'SERIAL':
        target.append(target_path + 'fmopt_mpi_fake.o')
    elif env['FMOPT_MODE'] == 'MPI':
        target.append(target_path + 'fmopt_mpi_interface.o')

    # Dependencies: These two files are always the same
    source.extend([target_path + 'fmopt_shared.global',
                   target_path + 'fmopt_user.global'])
    # This requires the naming conventions detailed above
    for suffix in ['.start', '.options']:
        source.append(target_path + target_body + suffix)
    # Find the fmopt directory and add its contents as dependencies
    fmpath = getfmpath(env.fs.pathTop)
    source.extend(Glob(fmpath + os.path.sep + '*?.*'))

    return target, source


def compilefmopt(target, source, env):
    """
    This function contains the sequence of steps to compile fmopt programs.
    Assumes that target prefix supplied is the same as the prefix of the main 
    source file. I.e. specifying: 'target = foo' assumes that the main source 
    file is 'foo.c'
    
    """

    # Get path to fmopt
    fmpath = getfmpath(env.fs.pathTop)

    # Get path to source = target files
    target_path = os.path.split(target[0].abspath)[0] + os.path.sep

    # Get sources in model directory to link them in
    srcfiles = ''
    for src in source:
        if src.abspath.startswith(target_path):
            if os.path.splitext(src.name)[1] == '.c':
                srcfiles = srcfiles + ' ' + src.name
                
    # Specify debug flags for gpc
    DEBUGFLAG = ''
    if env['DEBUG']:
        DEBUGFLAG = '-g'

    # Compile FMOPT's MPI interface if compiling for for parallel execution
    if env['FMOPT_MODE'] == 'MPI':
        stage = [(' '.join(['gpc', '-c',
                            '-D', 'OPENMPI',
                            '-D', env['FMOPT_MODE'],
                            '--automake',
                            os.path.join(fmpath, 'fmopt_mpi_interface.c')]))]
# These lines seem unnecessary.
#                            ,
#                            '-lgpc',
#                            '-lm',
#                            '-lmpi']))]
    elif env['FMOPT_MODE'] == 'SERIAL':
        stage = [None]

    # Compile the fmopt.pas file.
    stage.append(' '.join(['gpc', DEBUGFLAG, '-c',
                           '-D', 'OPENMPI',
                           '-D', env['FMOPT_MODE'],
                           '--automake', os.path.join(fmpath, 'fmopt.pas'),
                           '--unit-path=' + fmpath,
                           '-I' + fmpath,
                           '-I' + target_path ]))

    # Compile the main source file.
    stage.append(' '.join(['gpc', DEBUGFLAG, '-c',
                           '-D', 'OPENMPI',
                           '-D', env['FMOPT_MODE'],
                           '--automake', os.path.join(target_path, srcfiles),
                           '-I' + fmpath,
                           '--unit-path=' + fmpath,
                           '-I' + target_path ]))
    # Link the previous files. 
    o_files = []
    for tgt in target:
        if os.path.splitext(tgt.name)[1] == '.o':
            o_files.append(tgt.name)
    # Need to add some files from sources because SCons throws an error if they
    # are added as targets above (those already built for tests).
    for src in source:
        src_body, src_ext = os.path.splitext(src.name)
        if src_ext == '.c':
            # Only use files in target directory (else gives an error with mpi /
            # mpi fake -- whatever is not used).
            if os.path.split(src.path)[0] == os.path.split(target[0].path)[0]:
                tgt_name = src_body + '.o'
                if tgt_name not in o_files:
                    o_files.append(tgt_name)

    stage.append(' '.join(['gpc', DEBUGFLAG, 
                           '-o', target[0].name,
                           '-lgpc',
                           '-lm',
                           '-lmpi',
                           '-lgsl',
                           '-lgslcblas',
                           ' '.join(o_files)]))
    # Run the actual compilation commands
    for cmd in stage:
        if cmd:
            print cmd
            return_code = os.system(cmd)
            assert return_code == 0, '''\n
            Error compiling FMOPT while executing:\n\n''' + cmd + '\n\n\n'
    return


def getfmpath(topdir):
    '''
    Find path to fmopt directory by walking the directory tree defined by 
    topdir and looking for the fmopt.pas file, raise error if not existent.
    
    '''

    for (dirname, dirshere, fileshere) in os.walk(topdir):
        for filename in fileshere:
            if filename == 'fmopt.pas':
                fmpath = dirname
    assert os.path.exists(fmpath), \
             'No file fmopt.pas in directory tree' + topdir
    return fmpath


def exists(env):
    return 1
