__doc__ = \
"""
Contains SCons and Python functions to run a series of do-files, 
add this builder method to the SCons environment. Catches implicit
targets and dependencies by default. 
"""

__revision__ = '$Id: stata_builder.py 104 2009-11-23 14:13:39Z hgaudecker $'


from SCons.Script import *
from stata_emitter_scanner import dofile_scanner, dofile_emitter
from stata_options import determine_options


### @export "generate_env"
def generate(env):
    "add the builder and scanner from stata .do files to env"
    
    # define the builder (calls python functions rundofiles / printfiles)
    x = Action(rundofiles, printfiles)
    statabuilder = Builder(action = x, src_suffix = '.do', 
                           suffix = '.log', emitter = dofile_emitter)
    # add builder to the environment
    env.Append(BUILDERS = {'Stata' : statabuilder})
    
    # define scanner (actually only gets dependencies list from emitter)
    doscan = Scanner(function = dofile_scanner)
    # add scanner to the environment
    env.Append(SCANNERS = doscan)
### @end


def exists(env):
    return 1


### @export "rundofiles"
def rundofiles(target, source, env):
    """
    Python function running a series of do files from their source directories.
    Includes two checks whether build worked correctly: 
     1) File-based return code (stata file needs to call scons_returncode 
        at the end), checks whether EXIT0.STATA exists after the run.
     2) Check whether target files exist after the run (this suffices as SCons 
        removes existing targets if they need to be built).
        
    """
    from subprocess import call
    # extract filenames from SCons node objects
    # store current path, will need to change back to this 
    rootpath = os.getcwd()
    # set returncode to zero and change if anything goes wrong
    returncode = 0
    for source_node in source:
        # go to source directory of current do file
        path, source_file_name = os.path.split(str(source_node))
        os.chdir(path)
        # remove returncode file if still exists
        if os.path.exists('EXIT0.STATA'):
            os.remove('EXIT0.STATA')
        # call stata in batch mode, differs according to platform
        if sys.platform=='darwin':
            # launches whatever is found first of this list
            for cand in ['Stata64MP', 'StataMP', 'Stata64SE', 'StataSE', 'Stata64', 'Stata']:
                test = call(['which', '-s', cand])
                if test == 0:
                    statacmd = cand
                    break
                elif cand == 'Stata':
                    print_execmac_error()
                    returncode = 1
            # launch preferred stata command with appropriate flags for MacOS
            if returncode == 0:
                call([statacmd, '-e', '-q', 'do', source_file_name])
        elif sys.platform[:5]=='linux':
            call(['stata', '-b', '-q', 'do', source_file_name])
        
        # remove Stata's "return code", print error message if file not found.
        try:
            os.remove('EXIT0.STATA')
        except:
            print_run_error(str(source_node))
            returncode = 1
        
        # go back to original directory with SConscript file
        os.chdir(rootpath)
        
    if returncode is 0:
        # check whether targets have been built
        for target_node in target:
            if os.path.exists(str(target_node))==False:
                print_file_error(source, target_node)
                returncode = 1
        # remove standard log-file if desired, make sure that no error 
        # ocurred and that file exists (could be removed by do-file).
        for source_node in source:
            options = determine_options(target, source_node, env)
            if (options['log_delete']==True and 
                returncode == 0 and
                os.path.exists(str(options['default_log'])))==True:
                os.remove(str(options['default_log']))
    return returncode
### @end


def printfiles(target, source, env):
    """
    Prints message stating which do-files are run 
    and which targets are built by Stata builder.
    """
    printstring = 'Running Stata:'
    for src in source:
        printstring = printstring + '\n  ' + str(src)
    printstring = printstring + '\n' + 'to build:'
    for tgt in target:
        printstring = printstring + '\n  ' + str(tgt)
    printstring = printstring + '\n\n'
    return printstring


def print_execmac_error():
    sys.stderr.write("""
Error: Cannot find Stata executable on the system path. Looking
for StataMP / StataSE / stata. Make sure that the $PATH variable 
in your .bash_profile file includes something like: 
/Applications/Stata/Stata.app/Contents/MacOS and that there is a 
file with one of the three names above in it. Check by typing 
"which StataMP" (etc.) in a terminal window.
""")
    return


def print_run_error(source_node):
    sys.stderr.write("""
Run of '%s' did not finish properly
or it is missing the 'scons_returncode' call at the end\n
""" %str(source_node))
    return


def print_file_error(source, target_node):
    sys.stderr.write("""
Run of '%s'
did not produce target '%s'\n
""" %([str(x) for x in source], str(target_node)))
    return


