#!/usr/bin/env python
__doc__ = """
Contains functions to clean up Stata do-files. Can be called either from the
command line or from other python functions. The options are largely 
customisable. By default, the delimiter is changed to cr everywhere, asterisk-
style comments are replaced by double slashes, extra asterisks in comments are
removed, loops and conditional statements are indented by four spaces, 
continuation lines by two spaces, more than two consecutive blank lines are 
reduced to two.

IMPORTANT: The code is certain to contain bugs which might lead to broken 
Stata programs. Some issues are known, please read the accompanying 
documentation carefully. No guarantee whatsoever comes with this code.
"""

__revision__ = '$Id: stata_cleanup.py 52 2008-12-12 16:05:12Z hgaudecker $'

from pyparsing import Literal, Keyword, lineEnd, Suppress, Forward,\
    MatchFirst, SkipTo, lineStart, replaceWith, restOfLine,\
    originalTextFor, Optional, OneOrMore, Or, White, empty, CharsNotIn,\
    nestedExpr, Word, printables, ZeroOrMore, alphanums, Regex
from stata_expressions import strip_white, stata_quote, stata_comments,\
    control_flow_loop, program_def
import re, sys

global ichar, ino

### @export "cleanup_master"
def cleanup_master(**kwargs):
    """
    This function acts largely like a batch file, first retrieving
    the options and then calling functions to perform each cleaning step
    in turn. 
    """
    global ichar, ino
    
    options = get_options(kwargs)
    
    # put } in */} on a single line
    workstring = re.sub('\*/[ \t]*}', '*/\n}', options['instring'])
    
    # first strip all trailing and leading whitespace
    workstring = strip_white(workstring)
    
    # remove empty comments (will do so later again, but else the line end
    # converter has problems with multiple continued lines, i.e. 
    # display \\\ \n \\\ \n "hello" would produce an empty line in between)
    workstring = rmemptycomments(workstring)
    
    # conversion of line endings
    if options['convdelimit'] is True:
        workstring = convdelimit(workstring)
    
    # conversion of comments
    if options['convcomment'] is True:
        workstring = convcomments(workstring)
    
    # remove extra asterisks in comments
    if options['rmasterisks'] is True:
        workstring = rmasterisks(workstring)
    
    # remove extra blank spaces (do this before indentation!!!)
    if options['rmextrablank'] is True:
        workstring = re.sub('[ \t][ \t]*', ' ', workstring)
        
    # remove empty comments (usually redundant, 
    workstring = rmemptycomments(workstring)
    
    # strip all trailing and leading whitespace, just to be sure
    workstring = strip_white(workstring)
    
    # set number of multiple blank lines to defined value
    workstring = re.sub('\n{'+str(1+options['nblanklines'])+'}\n+', 
                        '\n'*(1+options['nblanklines']),
                        workstring)
    
    # put } in */} on a single line
    workstring = re.sub('\*/[ \t]*}', '*/\n}', workstring)
    
    # indentation of loops / if statements
    ichar = options['indlooptype']
    ino = options['indloopn']
    workstring = indent(workstring, 'loop')
    
    # indentation of continued lines
    ichar = options['indlctype']
    ino = options['indlcn']
    workstring = indent(workstring, 'lc')
    
    # indentation of programs
    ichar = options['indprogtype']
    ino = options['indprogn']
    workstring = indent(workstring, 'prog')
    
    # make comments prettier (insert blank after // and /// if not followed
    # by space or tab already)
    workstring = re.sub('//(?![\s/])', '// ', workstring)
    workstring = re.sub('///(?!\s)', '/// ', workstring)
    
    # determine output type depending on input type
    returnstring = None
    if 'file' in kwargs:
        f = open(options['file'], 'w')
        f.write(workstring)
        f.close()
    elif 'instring' not in kwargs:
        sys.stdout.write(workstring)
    else:
        returnstring = workstring
        
    return returnstring
### @end


### @export "get_options"
def get_options(kwargs):
    # determine whether instring exists or is to be read from stdin / file
    if 'instring' not in kwargs:
        if 'file' not in kwargs:
            kwargs['instring'] = sys.stdin.read()
        else:
            f = open(kwargs['file'])
            kwargs['instring'] = f.read()
            f.close()
    
    # conversion of the delimiter ; -> cr true 
    if 'convdelimit' in kwargs:
        try:
            kwargs['convdelimit'] = bool(int(kwargs['convdelimit']))
        except:
            raise Exception, "Specify: --convdelimit=[0][1]"
    else:
        kwargs['convdelimit'] = True
    
    # conversion of comments from * to //
    if 'convcomment' in kwargs:
        try:
            kwargs['convcomment'] = bool(int(kwargs['convcomment']))
        except:
            raise Exception, "Specify: --convcomment=[0][1]"
    else:
        kwargs['convcomment'] = True
    
    # remove extra asterisks in comments
    if 'rmasterisks' in kwargs:
        try:
            kwargs['rmasterisks'] = bool(int(kwargs['rmasterisks']))
        except:
            raise Exception, "Specify: --rmasterisks=[0][1]"
    else:
        kwargs['rmasterisks'] = True
    
    # whether indentation of loops / if statements as tabs or spaces
    if 'indlooptype' in kwargs:
        if kwargs['indlooptype'] == 'space':
            kwargs['indlooptype'] = ' '
        elif kwargs['indlooptype'] == 'tab':
            kwargs['indlooptype'] = '\t'
        else:
            raise Exception, "Specify: --indlooptype=[space][tab]"
    else:
        kwargs['indlooptype'] = ' '
    
    # number of tabs/spaces to indent for loops / if statements
    if 'indloopn' in kwargs:
        try:
            kwargs['indloopn'] = int(kwargs['indloopn'])
        except:
            raise Exception, "Specify: --indloopn=[integer]"
    else:
        kwargs['indloopn'] = 4
    
    # whether indentation after line continuation characters as tabs or spaces
    if 'indlctype' in kwargs:
        if kwargs['indlctype'] == 'space':
            kwargs['indlctype'] = ' '
        elif kwargs['indlctype'] == 'tab':
            kwargs['indlctype'] = '\t'
        else:
            raise Exception, "Specify: --indlctype=[space][tab]"
    else:
        kwargs['indlctype'] = ' '
    
    # number of tabs/spaces to indent after line continuation characters
    if 'indlcn' in kwargs:
        try:
            kwargs['indlcn'] = int(kwargs['indlcn'])
        except:
            raise Exception, "Specify: --indlcn=[integer]"
    else:
        kwargs['indlcn'] = 2
    
    # whether indentation of programs as tabs or spaces
    if 'indprogtype' in kwargs:
        if kwargs['indprogtype'] == 'space':
            kwargs['indprogtype'] = ' '
        elif kwargs['indprogtype'] == 'tab':
            kwargs['indprogtype'] = '\t'
        else:
            raise Exception, "Specify: --indprogtype=[space][tab]"
    else:
        kwargs['indprogtype'] = ' '
    
    # number of tabs/spaces to indent for programs
    if 'indprogn' in kwargs:
        try:
            kwargs['indprogn'] = int(kwargs['indprogn'])
        except:
            raise Exception, "Specify: --indprogn=[integer]"
    else:
        kwargs['indprogn'] = 4
    
    if 'nblanklines' in kwargs:
        try:
            kwargs['nblanklines'] = int(kwargs['nblanklines'])
        except:
            raise Exception, "Specify: --nblanklines=[integer]"
    else:
        kwargs['nblanklines'] = 2
    # concatenate the string for the re.sub command later on right here
    #kwargs['nblanklines'] = '\n*'+str(1+kwargs['nblanklines'])
    
    if 'rmextrablank' in kwargs:
        try:
            kwargs['rmextrablank'] = bool(int(kwargs['rmextrablank']))
        except:
            raise Exception, "Specify: --rmextrablank=[0][1]"
    else:
        kwargs['rmextrablank'] = True
    
    return kwargs
### @end


### @export "convdelimit"
def convdelimit(instring):
    
    # define the expression to set delimiter
    set_delimit = Literal('#') + Literal('delimit') + restOfLine + lineStart
    
    # do transformations only if the delimiter is actually changed somewhere
    if any(set_delimit.scanString(instring)) is False:
        outstring = instring
    else:
        # define a placeholder for current delimiter
        expr_delimit = Forward()
        
        # global dictionary for delimiter, only
        delimit_dict = {}
        
        # parse action for macro definitions
        def process_set_delimit(s, l, t):
            # delimiter is ; for anything but cr, define dict with replacements
            if t[2].strip() != 'cr':
                delimit_dict[';\n'] = '\n'
                delimit_dict['\n'] = ' ///\n'
                delimit_dict[';'] = '\n'
                expr_delimit << MatchFirst(map(Literal, delimit_dict.keys()))
            # if delimiter is set to cr, clear dict
            else:
                delimit_dict.clear()
                expr_delimit << Forward()
            # return linebreak only instead of entire token string
            return '\n'
        
        # parse action to replace relevant expressions
        def process_delimit_ref(s, l, t):
            return delimit_dict[t[0]]
        
        # attach parse actions to expressions
        expr_delimit.setParseAction(process_delimit_ref)
        set_delimit.setParseAction(process_set_delimit)
        
        # combine delimit expression and delimit definition
        delimit = originalTextFor(stata_quote_comm()) | expr_delimit | set_delimit
        
        # transform the string
        outstring = delimit.transformString(instring)
        outstring = strip_white(outstring)
        
        # post-transformation 1: remove /// from empty lines
        emptyline = Suppress(lineStart + Literal('///') + lineEnd)
        emptyline.setParseAction(replaceWith('\n'))
        outstring = emptyline.transformString(outstring)
        
        # post-transformation 2: remove /// before empty lines
        emptylinecont = Suppress(Literal('///') + lineStart + lineEnd)
        emptylinecont.setParseAction(replaceWith('\n\n'))
        outstring = emptylinecont.transformString(outstring)
        
        # post-transformation 3: remove ; and ; /// from lines with comments
        commwrong = (lineStart + (Regex('\*[^/]') ^ Literal('//')) + 
                        SkipTo(Literal(';'), include=False) +
                        Suppress(Literal(';') + Optional(Literal('///'))) + lineEnd)
        outstring = commwrong.transformString(outstring)
        
        # post-transformation 4: make sure } is on empty line (could be *\})
        outstring = re.sub('(?<=[^\n])\}', '\n}', outstring)
    
    return outstring
### @end


### @export "convcomments"
def convcomments(instring):
    def transform_astslash(s, l, t):
        # define some variables to make code simpler to read 
        cont = s[t[0]+2:t[-1]-2]
        full = s[t[0]:t[-1]]
        nlines = cont.count('\n')
        if nlines == 0:
            # check whether character(s) after */ is a line break
            check = re.match('^\n*', s[t[-1]:])
            # convert /* */ at the end of lines to //
            if check is not None:
                tnew = ' // ' + cont + '\n'*check.end()
            # convert /* */ in the middle of lines to ///
            else:
                tnew = ' /// ' + cont + '\n'
        elif nlines == 1:
            a = re.match('.*\n$', cont)
            # if there is something on the new line, leave comment as is
            if a is None:
                tnew = full
            # otherwise transform the string
            else:
                tnew = ' /// ' + cont
        else:
            tnew = full
        return tnew
    # conversion of comments from * to //
    astcomm = lineStart + Regex('\*[^/]').setParseAction(replaceWith('// '))
    outstring = astcomm.transformString(instring)
    # conversion of line continuations with /* ... \n */ to /// ... \n
    astslash_cont = originalTextFor(nestedExpr('/*', '*/', None, stata_quote()))
    astslash_cont.setParseAction(transform_astslash)
    outstring = astslash_cont.transformString(outstring)
    # put // on new line if not there already
    outstring = re.sub('(?<![\n/])//(?![/])', '\n//', outstring)

    
    return strip_white(outstring)
### @end


### @export "rmasterisks"
def rmasterisks(instring):
    # remove lines with extra asterisks at start of comments. 
    xtraast_beg = (Literal('//').setParseAction(replaceWith('// ')) + 
                   Suppress(OneOrMore(Literal('*'))))
    xtraast = (originalTextFor(stata_comments()['ast_slash_nested']) |
               originalTextFor(stata_comments()['ast_slash_eostring']) |
               xtraast_beg)
    outstring = xtraast.transformString(instring)
    return outstring
### @end


### @export "indent"
def indent(instring, type):
    if ino == 0:
        outstring = instring
    else:
        if type=='loop':
            outstring = control_flow_parse().transformString(instring)
        elif type=='lc':
            lc = (Literal('///') + (Suppress(lineEnd) | 
                                    originalTextFor(restOfLine)) + 
                  Suppress(lineStart) +
                  empty.setParseAction(replaceWith('\n'+ichar*ino)) + 
                  Optional(White()) + 
                  CharsNotIn('{', exact=1))
            outstring = lc.transformString(instring)
        elif type=='prog':
            program = (originalTextFor(stata_quote_comm()) | 
                       program_def().setParseAction(indent_program))
            outstring = program.transformString(instring)
    return outstring
### @end


### @export "indent_program"
def indent_program(s, l, t):
    # remove extra newlines at the end of program content
    t[2] = re.sub('\n*$', '', t[2])
    # add indentation level
    t[2] = re.sub('\n','\n'+ichar*ino, t[2])
    # put all pieces together
    tnew = ' '.join(t[0:2]) + t[2] + '\n' + ' '.join(t[3:4])
    return tnew
### @end


### @export "indent_cf"
def indent_cf(s, l, t):
    # loop over these nasty expressions because multiple ones 
    # in if-statement make dictionaries infeasible.
    control_flow = t._ParseResults__tokdict['control_flow']
    expr = t._ParseResults__tokdict['expr']
    content = t._ParseResults__tokdict['content']
    tnew = ''
    j = 0
    for cont in content:
        cont = cont[0]
        # do this recursively if any nested control flows are found
        if any(control_flow_check().scanString(cont)) is True:
            cont = control_flow_parse().transformString(cont)
        # replace linebreaks by linebreaks plus indent
        cont = cont.replace('\n', '\n'+ichar*ino)
        # undo this for the last lines
        cont = cont[:-2*(ino+1)]+'}\n'
        # take care of else if (concatenate string values):
        if type(control_flow[j][0]) is str:
            control_flow_str = control_flow[j][0]
        else:
            control_flow_str = ' '.join(control_flow[j][0])
        # take care of else (no expr statement)
        if control_flow_str == 'else':
            expression = ''
        else:
            expression = expr[j][0]
        # put everything back together
        tnew = tnew + (control_flow_str + ' ' + 
                       expression + cont)
        j += 1
    # deal with cases where no whitespace before opening brace
    tnew = re.sub('(?<=[^\s])\{', ' {', tnew)
    return tnew
### @end


### @export "rmemptycomments"
def rmemptycomments(instring):
    # empty double-slash comments: leave line there
    empty_dblslash = Suppress(lineStart + Keyword('//') + lineEnd)
    # empty line continuation: eat entire line
    empty_trplslash = Suppress(lineStart + Keyword('///') + lineEnd + lineStart)
    # transform string
    outstring = (empty_dblslash ^ empty_trplslash).transformString(instring)
    return outstring
### @end


### @export "control_flow_check"
def control_flow_check():
    # control flow without the continue statement
    control_flow = control_flow_loop()
    del control_flow['continue']
    control_flow = Or(control_flow.values())
    control_flow = originalTextFor(stata_quote_comm()) | control_flow
    return control_flow
### @end


### @export "control_flow_parse"
def control_flow_parse():
    # control flow including parse action and ignoring comments.
    control_flow = control_flow_loop()
    del control_flow['continue']
    control_flow = Or(control_flow.values())
    control_flow.setParseAction(indent_cf)
    control_flow = originalTextFor(stata_quote_comm()) | control_flow
    return control_flow
### @end


### @export "stata_quote_comm"
def stata_quote_comm():
    # returns expression with original string for stata quotes/comments
    expr = originalTextFor(stata_quote())
    for key in stata_comments().keys():
        expr = expr | originalTextFor(stata_comments()[key])
    return expr
### @end


### @export "stata_cleanup__main__"
if __name__ == "__main__":
    options = {}
    j = 1
    while j < len(sys.argv):
        if sys.argv[j][0:2] == '--':
            temp = sys.argv[j][2:].split('=')
            options[temp[0]] = temp[1]
        else:
            raise Exception, "Specify: options as --option=value"
        j+=1
    for key in options:
        if (key != 'file' and
            key != 'instring' and
            key != 'convdelimit' and 
            key != 'convcomment' and
            key != 'rmasterisks' and
            key != 'indlooptype' and
            key != 'indloopn' and
            key != 'indlctype' and
            key != 'indlcn' and
            key != 'indprogtype' and
            key != 'indprogn' and
            key != 'nblanklines' and
            key != 'rmextrablank'):
            raise Exception, """
Invalid option specified:  '%s' '%s' 
stata_cleanup.py takes the following options:
--file=[filename]
--instring=["string containing stata commands"]
--convdelimit=[0][1]
--convcomment=[0][1]
--rmasterisks=[0][1]
--indlooptype=[space][tab]
--indloopn=[integer]
--indlctype=[space][tab]
--indlcn=[integer]
--indprogtype=[space][tab]
--indprogn=[integer]
--nblanklines=[integer]
--rmextrablank=[0][1]
""" %(key, options[key])
    
    cleanup_master(**options)
### @end


