__doc__ = \
"""
Contains functions to scan a SCons node instance containing 
Stata do-files for a list of implicit targets and dependencies. 
"""
import re
__revision__ = '$Id: stata_emitter_scanner.py 78 2009-05-26 11:38:08Z hgaudecker $'

from SCons.Script import *
from pyparsing import Literal, replaceWith, lineStart, dblQuotedString, \
    nestedExpr, SkipTo, Or, lineEnd, Suppress, Word, alphas, alphanums, OneOrMore, \
    originalTextFor, CharsNotIn, nums, srange, Regex
from stata_expressions import strip_white, stata_quote, prepend, \
    control_flow_line, stata_comments, control_flow_loop, macros_def_only, \
    dep_def_only, tgt_def_only, abbrev_key, dep_tgt_ext, strip_stata_quotes
from stata_options import determine_options
from stata_cleanup import convdelimit


### @export "process_for_loop"
def process_for_loop(s, l, t):
    """
    Gets called when the parser has encountered a forvalues or foreach
    loop. First evaluates the expression after the loop statement, 
    obtaining a list of values for the looping variable. Then simulates
    the loop, updating the locals dictionary at each iteration, evaluating 
    token by calling all(parse_keywords.scanString(t.content))
    """
    global cont_break
    
    loop_local_name, loop_values = process_for_loop_helper(t.control_flow, t.expr)
    if loop_local_name in temp_macros:
        del(temp_macros[loop_local_name])
    for val in loop_values:
        local_macros[loop_local_name] = val
        all(parse_keywords.scanString(t.content))
        del local_macros[loop_local_name]
        if cont_break == True:
            cont_break = False
            break
    return
### @end


### @export "process_while"
def process_while(s, l, t):
    """
    Gets called when the parser has encountered a while-statement. 
    First evaluates the condition and then simulates the loop, updating 
    the locals dictionary at each iteration, evaluating the token by 
    calling all(parse_keywords.scanString(t.content))
    """
    global cont_break
    
    # expand macro, take care of potential Python/Stata 
    # difference in defining not equal
    try:
        decider = eval(macro_eval(t.expr, 
                           warn_on_unresolved=True, 
                           eval_on_def=False
                           ).replace('~','!'))
        while decider is True:
            all(parse_keywords.scanString(t.content))
            if cont_break is True:
                cont_break = False
                break
            else:
                decider = eval(macro_eval
                               (t.expr,
                                warn_on_unresolved=True,
                                eval_on_def=False
                                ).replace('~','!'))
    except:
        warn_eval_error_while(macro_eval
                              (t.expr,
                               warn_on_unresolved=True,
                               eval_on_def=False))
    return
### @end


### @export "process_if"
def process_if(s, l, t):
    """
    Gets called when the parser has encountered an if - statement. 
    First evaluates the expression(s), then relevant token.
    """
    # sets of tokens are not so easily accessible, rather work
    # with simpler names...
    process_if_core(t._ParseResults__tokdict['control_flow'],
                    t._ParseResults__tokdict['expr'],
                    t._ParseResults__tokdict['content'])
    return
### @end


### @export "process_continue"
def process_continue(s, l, t):
    """
    Gets called when the parser has encountered a continue [, break] 
    statement. Does not have to do anything except for setting the
    global variable cont_break if a break statement was present.
    """
    global cont_break
    
    if 'break' in t:
        cont_break = True
    return
### @end


### @export "process_macro"
def process_macro(s, l, t):
    """
    Gets called when the parser has encountered an command
    assigning a (new) value to a global, local, or temp.
    """
    if 'local' in t:
        process_local_def(t)
    elif 'glob' in t:
        process_global_def(t)
    elif 'temp' in t:
        process_temp_def(t)
    # call some function that parses the macro definition 
    # depending on whether it contains a = : or nothing
    # then update globals / locals / temps dictionaries.
    return
### @end


### @export "process_dep"
def process_dep(s, l, t):
    """
    Gets called when the parser has encountered an command
    defining a dependency.
    """
    global implicit_source_set, local_macros, temp_macros, current_do_file
    
    file_names = process_file_name(t.keys(), t.file)
    if file_names is not None:
        [implicit_source_set.add(x) for x in file_names]
        if 'include' in t:
            # replace current_do_file for warning messages
            tmp_dofile = current_do_file
            current_do_file = file_names[0]
            # process file
            file = open(file_names[0])
            include_string = file.read()
            get_deps_targets(include_string,
                             reset_locals=False,
                             reset_globals=False)
            # reset current_do_file to higher level one
            current_do_file = tmp_dofile
        elif 'dorun' in t:
            # do/run can be called with arguments, make sure that 
            # these are not included as filenames
            [implicit_source_set.remove(x) for x in file_names]
            implicit_source_set.add(file_names[0])
            # replace current_do_file for warning messages
            tmp_dofile = current_do_file
            current_do_file = file_names[0]
            # store local/temp macros since their scope does not go 
            # into the nested file called by do/run. 
            loc_mac = local_macros.items()
            tmp_mac = temp_macros.items()
            # process file
            file = open(file_names[0])
            include_string = file.read()
            get_deps_targets(include_string,
                             reset_locals=True,
                             reset_globals=False)
            # reset local/temp macros to what they are in original file
            local_macros.clear()
            temp_macros.clear()
            for item in loc_mac:
                local_macros[item[0]] = item[1]
            for item in tmp_mac:
                temp_macros[item[0]] = item[1]
            # reset current_do_file to higher level one
            current_do_file = tmp_dofile
    return
### @end


### @export "process_tgt"
def process_tgt(s, l, t):
    """
    Gets called when the parser has encountered an command
    defining a target.
    """
    global implicit_target_set
    
    if 'erase' not in t:
        file_names = process_file_name(t.keys(), t.file)
        if file_names is not None:
            [implicit_target_set.add(x) for x in file_names]
    else:
        file_name = process_file_name(t.keys(), t.file)[0]
        if file_name in implicit_target_set:
            implicit_target_set.remove(file_name)
    return
### @end


### @export "global_variables"
# sets for dependencies and targets
implicit_source_set = set()
implicit_target_set = set()

# store name of current do-file for warning messages
current_do_file = None

# dictionaries for defined macros
global_macros = {}
local_macros = {}
temp_macros = {}

# whether to break out of a loop
cont_break = False

# expressions for control_flow statements
control_flow = control_flow_loop()
control_flow['forvalues'].setParseAction(process_for_loop)
control_flow['foreach'].setParseAction(process_for_loop)
control_flow['while'].setParseAction(process_while)
control_flow['ifcmd'].setParseAction(process_if)
control_flow['continue'].setParseAction(process_continue)
control_flow_expr = Or(control_flow.values())

# expressions for Stata's globals / locals / temps
macros_def = macros_def_only()
for key in macros_def.keys():
    macros_def[key].setParseAction(process_macro)
macros_def_expr = Or(macros_def.values())

# expressions for commands generating targets / dependencies
dep_def = dep_def_only()
for key in dep_def.keys():
    dep_def[key].setParseAction(process_dep)
dep_def_expr = Or(dep_def.values())

tgt_def = tgt_def_only()
for key in tgt_def.keys():
    tgt_def[key].setParseAction(process_tgt)
tgt_def_expr = Or(tgt_def.values())

# put together all expressions for (second round) parsing
parse_keywords = (control_flow_expr | macros_def_expr | 
                  dep_def_expr | tgt_def_expr)
### @end


### @export "dofile_scanner"
def dofile_scanner(node, env, path):
    """
    Extracts list with implicit dependencies from node.attributes: this
    is being set by dofile_emitter earlier. If this has not happened
    or there are no implicit dependencies, it returns an empty list.
    """
    try:
        dependencies = node.attributes.dependency_list
    except:
        dependencies = []
    return dependencies
### @end


### @export "dofile_emitter"
def dofile_emitter(target, source, env):
    """
    Scans a SCons node instance containing Stata do-files for 
    a list of implicit targets and dependencies. Targets are added
    to SCons target instance, dependencies are stored in 
    node.attributes.dependency_list and used by stata_scanner()
    during build. SCons source instance is left unchanged.
    """
    global implicit_source_set, implicit_target_set, current_do_file
    
    for source_file_node in source:
        current_do_file = str(source_file_node)
        options = determine_options(target, 
                                    str(source_file_node), env)
        
        if ((options['log_delete']==False) and
            (options['default_log'] not in target)):
            target.append(options['default_log'])
        
        if options['scan_file']==True:
            # scons function to read in the node's contents
            dofile = source_file_node.get_contents()
            # reset dependencies/targets at each iteration
            implicit_source_set.clear()
            implicit_target_set.clear()
            # create empty list of implicit dependencies
            source_file_node.attributes.dependency_list = []
            # core function extracting all dependencies/targets from node
            get_deps_targets(dofile)
            # convert implicit dependencies/targets to SCons nodes, sort
            source_implicit = [File(x) for x in implicit_source_set]
            source_implicit.sort()
            target_implicit = [File(x) for x in implicit_target_set]
            target_implicit.sort()
            # append implicit dependencies/targets if not present already
            for src in source_implicit:
                if src not in source:
                    source_file_node.attributes.dependency_list.append(src)
            for tgt in target_implicit:
                if tgt not in target:
                    target.append(tgt)
    return target, source
### @end


### @export "get_deps_targets"
def get_deps_targets(dofile, reset_locals=True, reset_globals=True):
    """
    Core function extracting all dependencies and targets 
    from a given Stata do-file. By default, the dictionaries with 
    locals and globals are reset, change this behaviour depending
    on whether the function is called for an explicit node (default),
    for a do/run implicit dependency (reset locals but not globals),
    or for an include implicit dependency (do not reset anything)
    """
    # reset dictionaries with macro names if desired
    if reset_locals is True:
        local_macros = {}
        temp_macros = {}
    if reset_globals is True:
        global_macros = {}
    # set delimiter to cr everywhere
    dofile = convdelimit(dofile)
    # get rid of all the comments in the do-file
    dofile = weed_out_comments(dofile)
    # remove all lines not relevant for input / output
    dofile_io = weed_out_unrelated(dofile)
    # parse time! 
    all(parse_keywords.scanString(dofile_io))
    
    return
### @end


### @export "weed_out_comments"
def weed_out_comments(dofile):
    """
    Cleans up Stata do-files: Removes all comments, strips leading
    and trailing whitespace, removes empty lines.
    """
    # convert line endings and strip extra whitespace
    dofile = strip_white(dofile)
    
    # get dict with comments from module stata_expressions, 
    # remove them sequentially (cannot handle all of them jointly)
    comments = stata_comments()
    comments_group = (comments['trpl_slash'] | 
                      comments['dbl_slash'] | 
                      comments['ast'] | 
                      comments['ast_slash_nested'])
    for c in [comments['ast_slash_ast'], 
                    comments_group, 
                    comments['ast_slash_eostring']]:
        c.setParseAction(replaceWith(''))
        c.ignore(stata_quote())
        dofile = c.transformString(dofile)
    
    # strip trailing and leading whitespace, remove empty lines
    dofile = strip_white(dofile, rmblanklines=True)
    
    return dofile
### @end


### @export "weed_out_unrelated"
def weed_out_unrelated(dofile):
    """
    Get rid of all lines without input/output statements, 
    macros, loops, if-statements
    """
    keywords = {}
    keywords.update(control_flow_line())
    keywords.update(macros_def_only())
    keywords.update(dep_def_only())
    keywords.update(tgt_def_only())
    keywords_expr = Or(keywords.values())
    
    #fill up an empty string with lines relevant for i/o.
    dofile_io = ''
    for line in dofile.split('\n'):
        try:
            keywords_expr.parseString(line)
        except:
            continue
        dofile_io = dofile_io + line + '\n'
    
    # remove empty control flow statements
    keywords = control_flow_line()
    del keywords['continue']
    del keywords['closebrace']
    keywords_expr = Or(keywords.values())
    empty_controlflow = Suppress(keywords_expr +
                                 Literal('}').ignore(stata_quote()))
    while any(empty_controlflow.scanString(dofile_io)) is True:
        dofile_io = empty_controlflow.transformString(dofile_io)
    
    # strip trailing and leading whitespace, remove empty lines
    dofile_io = strip_white(dofile_io, rmblanklines=True)
    
    return dofile_io
### @end


### @export "process_for_loop_helper"
def process_for_loop_helper(control_flow, raw_expr):
    """
    takes the loop statement (i.e. forvalues / foreach)
    and processes the expression containend in raw_expr
    """
    loop_local_expr = Word(alphas+'_', alphanums+'_'
                           ).setResultsName('loop_local')
    raw_expr = macro_eval(raw_expr, warn_on_unresolved=True)
    if control_flow.startswith('forv'):
        loop_local, loop_values = process_forvalues(control_flow, 
                                                    raw_expr, 
                                                    loop_local_expr)
    elif control_flow=='foreach':
        loop_local, loop_values = process_foreach(control_flow, 
                                                  raw_expr, 
                                                  loop_local_expr)
    return loop_local, loop_values
### @end


### @export "process_forvalues"
def process_forvalues(control_flow, raw_expr, loop_local_expr):
    str_real = Word(nums+srange('[.+-]'))
    startval_expr = str_real.setResultsName('startval')
    endval_expr = str_real.setResultsName('endval')
    incr_expr = str_real.setResultsName('incr')
    secval_expr = str_real.setResultsName('secval')
    pat1 = (loop_local_expr + Literal('=') + startval_expr +
            Literal('(') + incr_expr + Literal(')') + endval_expr)
    pat2 = (loop_local_expr + Literal('=') + startval_expr + 
            Literal('/') + endval_expr)
    pat3 = (loop_local_expr + Literal('=') + startval_expr +
            secval_expr + Or(Literal(':'), Literal('to')) + endval_expr)
    try:
        obj = pat1.parseString(raw_expr)
        incr = obj.incr
        incr_bool = eval("'.' in incr")
    except:
        try:
            obj = pat2.parseString(raw_expr)
            incr = '1'
            incr_bool = eval("'.' in incr")
        except:
            obj = pat3.parseString(raw_expr)
            incr = int(float(obj.secval)) - int(float(obj.firstval))
            incr_bool = eval("'.' in obj.secval" or "'.' in obj.firstval")
    
    if ('.' in obj.startval) or ('.' in obj.endval) or (incr_bool is True):
        warn_forvalues_integer(raw_expr)
        
    startval = int(float(obj.startval))
    endval = int(float(obj.endval))
    incr = int(float(incr))
    # if loop ends exactly at endval: add incr to range function 
    # in order to match python/stata loops.
    if ((endval-startval) % incr) is 0:
        endval += incr
    loop_values = range(startval, endval, incr)
    
    return obj.loop_local, loop_values
### @end


### @export "process_foreach"
def process_foreach(control_flow, raw_expr, loop_local_expr):
    placeholders = (Literal('*') ^ Literal('-') + 
                    Literal('/') ^ Literal('(') + 
                    Literal(')'))
    placeholders.setParseAction(replaceWith(" "))
    expr = placeholders.transformString(raw_expr)
    if expr is not raw_expr:
        warn_foreach_placeholder(raw_expr, expr)
        
    list_expr = (OneOrMore(Word(alphanums + srange("[`'_]")) ^ 
                          dblQuotedString).setResultsName('loop_values') +
                 Suppress(lineEnd))
    pat_in = loop_local_expr + Literal('in') + list_expr
    pat_of_local = (loop_local_expr + Literal('of') + 
                    abbrev_key('loc', 'local') + list_expr)
    pat_of_global = (loop_local_expr + Literal('of') +
                     abbrev_key('glo', 'global') + list_expr)
    pat_of_list =  (loop_local_expr + Literal('of') + 
                    (abbrev_key('var', 'varlist') |
                     abbrev_key('new', 'newlist') |
                     abbrev_key('num', 'numlist')) + list_expr)
    try:
        obj = pat_in.parseString(expr)
        loop_values = ' '.join(obj.loop_values)
    except:
        try:
            obj = pat_of_local.parseString(expr)
            loop_values = "`" + obj.loop_values[0] + "'"
            loop_values = macro_eval(loop_values, warn_on_unresolved=True)
        except:
            try:
                obj = pat_of_global.parseString(expr)
                loop_values = "${" + obj.loop_values[0] + "}"
                loop_values = macro_eval(loop_values, warn_on_unresolved=True)
            except:
                obj = pat_of_list.parseString(expr)
                loop_values = ' '.join(obj.loop_values)
    loop_values = loop_values.split()
    return obj.loop_local, loop_values
### @end


### @export "process_if_core"
def process_if_core(control_flow, expr, content):
    j = 0
    for statement in control_flow:
        if statement[0] is not 'else':
            # expand macros
            eval_expr = macro_eval(expr[j][0], 
                                   warn_on_unresolved=True, 
                                   eval_on_def=False)
            # take care of Python/Stata difference in defining not equal
            eval_expr = eval_expr.replace('~','!')
            try:
                decider = eval(eval_expr)
            except:
                warn_eval_error_if(eval_expr)
                decider = False
            if decider is True:
                all(parse_keywords.scanString(content[j][0]))
                break
            else:
                j += 1
        else:
            all(parse_keywords.scanString(content[j][0]))
            break
        
    return
### @end


### @export "process_global_def"
def process_global_def(t):
    """
    Processes the definition of a global macro, updating
    the dictionary of global macros.
    """
    global_macros[t.glob] = process_macro_core(t.value, t.glob, 'global')
    return
### @end


### @export "process_local_def"
def process_local_def(t):
    """
    Processes the definition of a local macro, updating
    the dictionary of local macros.
    """
    if 'add_sub' in t:
        if t.add_sub[0] is '+':
            local_macros[t.local] = int(local_macros[t.local]) + 1
        elif t.add_sub[0] is '-':
            local_macros[t.local] = int(local_macros[t.local]) - 1
    else:
        local_macros[t.local] = process_macro_core(t.value, t.local, 'local')
        if t.local in temp_macros:
            del(temp_macros[t.local])
    
    return
### @end


### @export "process_temp_def"
def process_temp_def(t):
    tempname_list = (t.temp).split()
    for tempname in tempname_list:
        tempname = macro_eval(tempname)
        temp_macros[tempname] = ''
        if tempname in local_macros:
            del(local_macros[tempname])
    return
### @end


### @export "process_macro_core"
def process_macro_core(core_expr, macname, type):
    if core_expr.strip()[0] is ':':
        macval = ''
        warn_ext_fct(type, macname, core_expr)
    else:
        if core_expr.strip()[0] is '=':
            # evaluate the expression after the equals sign,
            # note that only simple operations that have the 
            # same expressions in Stata and Python will be 
            # processed.
            eval_def = True
            core_expr = core_expr.strip()[1:]
        else:
            eval_def = False
        
        core_expr = strip_stata_quotes(core_expr)
        macval = macro_eval(core_expr,warn_on_unresolved=True,
                            eval_on_def=eval_def)
        
    return macval
### @end


### @export "process_file_name"
def process_file_name(keys, in_files):
    files = []
    if type(in_files) is str:
        files.append(in_files)
    else:
        files = in_files
    # should be only one key except for file, but this is simple...
    for key in keys:
        if key is not 'file':
            for i in range(len(files)):
                file = macro_eval(files[i], 
                                  warn_on_unresolved=True, 
                                  eval_on_def=False)
                # take care of case where entire file is in quotes
                file = strip_stata_quotes(file)
                filebase, fileext = os.path.splitext(file)
                # take care of case where only filename is in quotes
                filebase = strip_stata_quotes(filebase)
                if filebase != '':
                    if fileext is '':
                        fileext = dep_tgt_ext()[key]
                    file = filebase.strip() + fileext.strip()
                else:
                    # if no filename is given (likely to be a tempvar),
                    # assume that there is no outfile. Mimics behaviour
                    # of Stata except for cases like "save .dta" ... 
                    # These should not happen anyhow! 
                    file = None
                # remove Windows-style filenames
                if '\\' in file:
                    file = file.replace('\\','/')
                # remove Windows-style filenames
                if '"' in file:
                    file = file.replace('"','')
                files[i] = file
    if files == []:
        files = None
    return files
### @end


### @export "macro_eval"
def macro_eval(in_string, warn_on_unresolved=True, eval_on_def=False):
    """
    Evaluates references to Stata macros. If warn_on_unresolved
    is set to True, prints a a warning if a reference to a 
    macro cannot be resolved.
    """
    
    # replace macros in in_string by their values
    out_string = macro_ref(warn_on_unresolved).transformString(in_string)
    # iterate over this if some macros contained references 
    # to macros themselves (i.e. a statement like local A ${B}).
    while ("`" in out_string) or ("${" in out_string):
        out_string = macro_ref(warn_on_unresolved).transformString(out_string)
        
    if ('$' in out_string) and (warn_on_unresolved is True):
        warn_global_nobraces(in_string, out_string)
    
    # handle & and |
    out_string = re.sub('&', 'and', out_string)
    out_string = re.sub('\|', 'or', out_string)
    
    if eval_on_def is True:
        try:
            # Note: Can only evaluate statements that 
            # conform to python syntax, like simple additions
            # and so on. Can't think of a rule that a meaningful
            # error message could be issued since often strings
            # are involved, for example in a statement:
            # global A B
            # local c = "${A}"
            # But usually this should not be relevant for i/o
            out_string = str(eval(out_string))
        except:
            out_string = out_string
    return out_string
### @end


### @export "macro_ref"
def macro_ref(warn_on_unresolved):
    """ 
    Processes the references to macros. Looks up locals and 
    globals recursively in dictionary and returns the input 
    string with expanded macros, unchanged string if no 
    macros were contained in it. 
    """
    def local_lookup(s, l, t):
        "Looks up local macros in dictionaries"
        try:
            # recursive operation: find nested macros in token
            # for macro without ` and '
            macro = macro_ref(warn_on_unresolved
                              ).transformString(t[1][1:-1])
        except:
            # use macro expression without ` and '
            macro = t[1][1:-1]
            
        if macro[0]=='"' and macro[-1]=='"':
            # don't take compound double quotes for macros,
            # remove compound part and return double quoted string
            t[1] = macro
        else:
            # look up macros in dictionaries, use empty string 
            # otherwise and issue a warning if set.
            if macro in local_macros:
                t[1] = local_macros[macro]
            elif macro in temp_macros:
                t[1] = temp_macros[macro]
            else:
                if warn_on_unresolved is True:
                    warn_unresolved_macro(s, t[1])
                t[1] = ''
                
        return t
    
    def global_lookup(s, l, t):
        "Looks up global macros in dictionary"
        try:
            # recursive operation: find nested macros in token
            # for macro without ${ and }
            macro = macro_ref(warn_on_unresolved).transformString(t[1][2:-1])
        except:
            # use macro expression without ${ and }
            macro = t[1][2:-1]
            
        # look up macros in dictionary, use empty string 
        # otherwise and issue a warning if set.
        if macro in global_macros:
            t[1] = global_macros[macro]
        else:
            if warn_on_unresolved is True:
                warn_unresolved_macro(s, t[1])
            t[1] = ''
        return t
    
    # define expressions for local macros
    local_ref = (originalTextFor(SkipTo(Literal("`"))) + 
                 originalTextFor(nestedExpr("`", "'", None, None), 
                                 asString=True).setResultsName('local')
                 )
    local_ref.leaveWhitespace()
    local_ref.setParseAction(local_lookup)
    # define expressions for global macros
    global_ref = (originalTextFor(SkipTo(Literal("${"))) + 
                  originalTextFor(nestedExpr("${", "}", None, None), 
                                  asString=True).setResultsName('global')
                  )
    global_ref.leaveWhitespace()
    global_ref.setParseAction(global_lookup)
    # joint expression: find either global or local
    macro_ref = local_ref | global_ref
    return macro_ref
### @end


def warn_forvalues_integer(raw_expr):
    sys.stderr.write("""\n
Warning (Stata Emitter/Scanner), """ + current_do_file + """:
forvalues loop containing non-integers not handled by
Stata Emitter/Scanner, values converted to integers:
forvalues """ + raw_expr + """
If this is a problem for dependency / target names, 
replace Stata loop indices by integers or use 
foreach loop.\n""")


def warn_foreach_placeholder(raw_expr, expr):
    sys.stderr.write("""\n
Warning (Stata Emitter/Scanner), """ + current_do_file + """:
foreach loop contains placeholders (- *) or numlist
with / or (), but Stata Emitter/Scanner can't handle 
data-dependent loops or more implicit numlists: 
foreach """ + raw_expr + """  
has been converted to: 
foreach """ + expr + """ 
If this is a problem for dependency / target names,
use only explicit items in foreach loop.""")


def warn_unresolved_macro(full_expression, macro):
    sys.stderr.write("""\n
Warning (Stata Emitter/Scanner), """ + current_do_file + """: 
Cannot resolve macro: """ + macro + """,
has been set to emtpy string.\n""")


def warn_eval_error_if(expression):
    sys.stderr.write("""\n
Warning (Stata Emitter/Scanner), """ + current_do_file + """: 
Error evaluating expression: """ + expression + """,
in if / else if statement. Has been set to False.\n""")


def warn_eval_error_while(expression):
    sys.stderr.write("""\n
Warning (Stata Emitter/Scanner), """ + current_do_file + """: 
Error evaluating expression: """ + expression + """,
in while loop. Broke out of loop.\n""")


def warn_global_nobraces(in_string, out_string):
    sys.stderr.write("""\n
Warning (Stata Emitter/Scanner), """ + current_do_file + """: \n
Evaluated: """ + in_string + """ to equal:
""" + out_string + """. If the latter contains
a global macro that was referenced without curly braces, 
please use curly brace notation, i.e. reference 
a global macro XYZ by ${XYZ} and not $XYZ.
\n""")


def warn_ext_fct(mactype, macname, macval):
    sys.stderr.write("""\n\n
Warning (Stata Emitter/Scanner), """ + current_do_file + """: \n
Stata's extended functions are not supported.
Found """ + mactype + ' ' + macname + ' ' + macval + """,
set """ + mactype + ' ' + macname  + """ to empty string.
If this is a problem for dependency / target names, please 
use another way to load/create files.\n""")


