__doc__ = "pyparsing expressions for some Stata syntax."

__revision__ = '$Id: stata_expressions.py 46 2008-10-21 08:50:01Z hgaudecker $'

from pyparsing import restOfLine, Literal, lineStart, dblQuotedString, \
    dblSlashComment, nestedExpr, stringEnd, SkipTo, Or, lineEnd, ZeroOrMore, \
    Optional, Suppress, Word, alphas, alphanums, empty, Keyword, OneOrMore, Group, \
    originalTextFor, QuotedString, removeQuotes, White, Regex, CharsNotIn,\
    printables
import re


### @export "compound_double_quote"
def compound_double_quote():
    quote = QuotedString('`"', escChar=None, escQuote=None, 
                              multiline=False, unquoteResults=True, 
                              endQuoteChar=""""'""")
    return quote
### @end


### @export "stata_quote"
def stata_quote():
    "Defines Stata's quoted strings "
    dbl_quote_eol = Literal('"') + restOfLine
    stata_quote = dblQuotedString | compound_double_quote() | dbl_quote_eol
    return stata_quote
### @end


### @export "standard_constructs"
filename = (Word(alphanums+'_'+'-'+'.'+'`'+"'"+'"'+'$'+'{'+'}'+'/'+'\\') ^ 
            dblQuotedString.setParseAction(removeQuotes))
suppr_vars_using = Suppress(SkipTo(Keyword('using'), 
                                   include=True, 
                                   ignore=(stata_quote() | 
                                           lineEnd + SkipTo(stringEnd)
                                           )
                                   )
                            )
suppr_using = Suppress(Keyword('using'))
require_eol_or_opts = Suppress(Or([lineEnd, (Literal(',') + restOfLine)]))
### @end


### @export "stata_comments"
def stata_comments():
    "Returns a dictionary with Stata's comments"
    comments = {}
    comments['trpl_slash'] = Literal('///') + restOfLine + lineStart
    comments['dbl_slash'] = dblSlashComment
    comments['ast']= lineStart + Regex('\*[^/]') + restOfLine
    comments['ast_slash_nested'] = nestedExpr('/*','*/',None,None)
    comments['ast_slash_eostring'] = (Literal('/*') + 
                                      SkipTo(stringEnd, failOn=~Literal('*/')))
    comments['ast_slash_ast'] = Literal('/*/')
    return comments
### @end


### @export "control_flow_line"
def control_flow_line():
    """
    Defines the keywords for Stata's control flow statements 
    on a line-by-line basis in order to be able to keep those
    lines in a parser.
    """ 
    # Define empty dictionaries
    keywords = {}
    keywords['forvalues'] = (lineStart + prepend() + 
                             abbrev_key('forv', 'forvalues') + 
                             restOfLine)
    keywords['foreach'] = (lineStart + prepend() + 
                           Keyword('foreach') + restOfLine)
    keywords['while'] = (lineStart + prepend() + 
                         Keyword('while') + restOfLine)
    keywords['ifcmd'] = (lineStart + prepend() + 
                         Keyword('if') + restOfLine)
    keywords['elseif'] = (lineStart + prepend() + 
                          Keyword('else') + Keyword('if') + restOfLine)
    keywords['else'] = lineStart + prepend() + Keyword('else') + Literal('{')
    keywords['continue'] = (lineStart + prepend() + 
                            Keyword('continue') + restOfLine)
    keywords['closebrace'] = lineStart + '}' + lineEnd
    return keywords
### @end


### @export "control_flow_loop"
def control_flow_loop():
    """
    Defines the keywords for Stata's control flow statements 
    together with the following { ... } statement.
    """ 
    keywords = {}
    keywords['forvalues'] = (lineStart + prepend() +
                             abbrev_key('forv', 'forvalues'
                                        ).setResultsName('control_flow') + 
                             SkipTo('{').setResultsName('expr') + 
                             originalTextFor(nestedExpr('{', '}\n', None, 
                                                        stata_quote())
                                             ).setResultsName('content'))
    keywords['foreach'] = (lineStart + prepend() +
                           Keyword('foreach').setResultsName('control_flow') + 
                           SkipTo('{').setResultsName('expr') + 
                           originalTextFor(nestedExpr('{', '}\n', None,
                                                      stata_quote())
                                           ).setResultsName('content'))
    keywords['while'] = (lineStart + prepend() +
                         Keyword('while').setResultsName('control_flow') + 
                         SkipTo('{').setResultsName('expr') + 
                         originalTextFor(nestedExpr('{', '}\n', None,
                                                    stata_quote())
                                         ).setResultsName('content'))
    keywords['ifcmd'] = (lineStart + prepend() +
                         Keyword('if').setResultsName('control_flow') + 
                         SkipTo('{').setResultsName('expr') + 
                         originalTextFor(nestedExpr('{', '}\n', None,
                                                    stata_quote())
                                         ).setResultsName('content') +
                         ZeroOrMore(lineStart + prepend() +
                                    Group(Keyword('else') +Keyword('if')
                                          ).setResultsName('control_flow') +
                                    SkipTo('{').setResultsName('expr') + 
                                    originalTextFor(nestedExpr('{', '}\n', None,
                                                               stata_quote())
                                                    ).setResultsName('content')
                                    ) + 
                         Optional(lineStart + prepend() + 
                                  Keyword('else'
                                          ).setResultsName('control_flow') +
                                  originalTextFor(nestedExpr('{', '}\n', None,
                                                             stata_quote())
                                                  ).setResultsName('content')
                                  )
                         )
    keywords['continue'] = (lineStart + prepend() + Keyword('continue') + 
                            Optional(Literal(',') + 
                                     abbrev_key('br', 'break'
                                                ).setResultsName('break')
                                     ) + SkipTo(stringEnd))
    return keywords
### @end


### @export "macros_def_only"
def macros_def_only():
    " Defines the keywords for Stata's macro definitions."
    keywords = {}
    keywords['local_def'] = (lineStart + prepend() +
                             Suppress(abbrev_key('loc', 'local')) +
                             Word(alphas+'_', alphanums+'_'
                                  ).setResultsName('local') +
                             restOfLine.setResultsName('value'))
    keywords['local_add_sub'] = (lineStart + prepend() +
                                 Suppress(abbrev_key('loc', 'local')) +
                                 (Literal('++') ^ Literal('--')
                                  ).setResultsName('add_sub') +
                                 restOfLine.setResultsName('local'))
    keywords['global_def'] = (lineStart + prepend() + 
                              Suppress(abbrev_key('gl', 'global')) +
                              Word(alphas, alphanums+'_'
                                   ).setResultsName('glob') +
                              restOfLine.setResultsName('value'))
    keywords['temp_def'] = (lineStart + prepend() +
                            Suppress(Keyword('tempvar') |
                                     Keyword('tempfile') |
                                     Keyword('tempname')) +
                            restOfLine.setResultsName('temp'))
    return keywords
### @end


### @export "dep_def_only"
def dep_def_only():
    "Defines the keywords for dependencies"
    filenamelist = OneOrMore(~lineEnd+filename)
    keywords = {}
    keywords['include'] = (lineStart + prepend() + 
                           Keyword('include').setResultsName('include') + 
                           filename.setResultsName('file'))
    keywords['dorun'] = (lineStart + prepend() +
                         (Keyword('do') | abbrev_key('ru', 'run')
                          ).setResultsName('dorun') +
                         filename.setResultsName('file'))
    keywords['use'] = (lineStart + prepend() + 
                       abbrev_key('u','use').setResultsName('use') +
                       (suppr_vars_using + filename.setResultsName('file') |
                        filename.setResultsName('file')) + require_eol_or_opts)
    keywords['fdause'] = (lineStart + prepend() + 
                          Keyword('fdause').setResultsName('fdause') +
                          filename.setResultsName('file') + require_eol_or_opts)
    keywords['xmluse'] = (lineStart + prepend() + 
                          Keyword('xmluse').setResultsName('xmluse') +
                          filename.setResultsName('file') + 
                          require_eol_or_opts)
    keywords['infile'] = (lineStart + prepend() +
                          abbrev_key('inf', 'infile').setResultsName('infile') +
                          suppr_vars_using + filename.setResultsName('file') +
                          require_eol_or_opts)
    keywords['insheet'] = (lineStart + prepend() +
                           Keyword('insheet').setResultsName('insheet') +
                           suppr_vars_using + filename.setResultsName('file') +
                           require_eol_or_opts)
    keywords['merge'] = (lineStart + prepend() +
                         (abbrev_key('mer', 'merge') |
                          Keyword('xtnmerge')
                          ).setResultsName('merge') +
                         (suppr_vars_using + 
                          filenamelist.setResultsName('file') |
                          filenamelist.setResultsName('file')
                          ) + require_eol_or_opts)
    keywords['append'] = (lineStart + prepend() +
                          abbrev_key('app', 'append').setResultsName('append') +
                          suppr_using + filename.setResultsName('file') +
                          require_eol_or_opts)
    keywords['joinby'] = (lineStart + prepend() +
                          Keyword('joinby').setResultsName('joinby') +
                          (suppr_vars_using + filename.setResultsName('file') |
                           filename.setResultsName('file')) + 
                          require_eol_or_opts)
    keywords['cross'] = (lineStart + prepend() +
                         Keyword('cross').setResultsName('cross') +
                         suppr_using + filename.setResultsName('file') +
                         require_eol_or_opts)
    keywords['fileread'] = (lineStart + prepend() +
                            Group(Keyword('file') + Keyword('open')
                                  ).setResultsName('fileread') +
                            Suppress(Word(alphas+'_', alphanums+'_')) + 
                            suppr_using + 
                            filename.setResultsName('file') + Suppress(',') +
                            Suppress(SkipTo(Keyword('read')) + restOfLine))
    return keywords
### @end


### @export "tgt_def_only"
def tgt_def_only():
    """
    Defines the keywords for targets and default 
    extensions by command.
    """ 
    keywords = {}
    keywords['save'] = (lineStart + prepend() + 
                        (abbrev_key('sa', 'save') ^ Keyword('saveold')
                         ).setResultsName('save') +
                         filename.setResultsName('file') + require_eol_or_opts)
    keywords['fdasave'] = (lineStart + prepend() + 
                           abbrev_key('fdasav', 'fdasave'
                                      ).setResultsName('fdasave') + 
                           (suppr_vars_using + filename.setResultsName('file') |
                            filename.setResultsName('file')) +
                            require_eol_or_opts)
    keywords['xmlsave'] = (lineStart + prepend() + 
                           abbrev_key('xmlsav', 'xmlsave'
                                      ).setResultsName('xmlsave') +
                           (suppr_vars_using + filename.setResultsName('file') |
                            filename.setResultsName('file')) +
                            require_eol_or_opts)
    keywords['outfile'] = (lineStart + prepend() +
                           abbrev_key('ou', 'outfile'
                                      ).setResultsName('outfile') +
                           suppr_vars_using + filename.setResultsName('file') + 
                           Suppress(restOfLine))
    keywords['outsheet'] = (lineStart + prepend() +
                            Keyword('outsheet').setResultsName('outsheet') +
                            suppr_vars_using + filename.setResultsName('file') + 
                            require_eol_or_opts)
    keywords['graph_export'] = (lineStart + prepend() + 
                                Group(abbrev_key('gr', 'graph') +
                                      Keyword('graph_export')
                                      ).setResultsName('graph_export') +
                                      filename.setResultsName('file') + 
                                      require_eol_or_opts)
    keywords['graph_export_pdf'] = (lineStart + prepend() + 
                                    Keyword('graph_export_pdf'
                                            ).setResultsName('graph_export_pdf')
                                    + filename.setResultsName('file') + 
                                    require_eol_or_opts)
    keywords['log'] = (lineStart + prepend() +
                       Group(Keyword('log') + Keyword('using')
                             ).setResultsName('log') + 
                       filename.setResultsName('file') + require_eol_or_opts)
    keywords['filewrite'] = (lineStart + prepend() + 
                             Group(Keyword('file') + Keyword('open')
                                   ).setResultsName('filewrite') +
                             Suppress(Word(alphas+'_', alphanums+'_')) + 
                             suppr_using + filename.setResultsName('file') + 
                             Suppress(',') + 
                             Suppress(SkipTo(Keyword('write'),True) + 
                                      restOfLine))
    # generic syntax for ... saving(filename[.dta], ...) ...
    keywords['saving_dta'] = (Literal('saving(').setResultsName('saving_dta') +
                              filename.setResultsName('file') + 
                              Suppress(restOfLine))
    keywords['erase'] = (lineStart + prepend() +
                         Keyword('erase').setResultsName('erase') +
                         filename.setResultsName('file') + Suppress(lineEnd))
    return keywords
### @end


### @export "dep_tgt_ext"
def dep_tgt_ext():
    """
    Defines default extensions for by keyword
    for dependencies and targets.
    """
    extensions = {}
    extensions['include'] = '.do'
    extensions['dorun'] = '.do'
    extensions['use'] = '.dta'
    extensions['fdause'] = '.xpt'
    extensions['xmluse'] = '.xml'
    extensions['infile'] = '.raw'
    extensions['insheet'] = '.raw'
    extensions['merge'] = '.dta'
    extensions['append'] = '.dta'
    extensions['joinby'] = '.dta'
    extensions['cross'] = '.dta'
    extensions['fileread'] = ''
    extensions['save'] = '.dta'
    extensions['fdasave'] = '.xpt'
    extensions['xmlsave'] = '.xml'
    extensions['outfile'] = '.raw'
    extensions['outsheet'] = '.raw'
    extensions['graph_export'] = ''
    extensions['graph_export_pdf'] = '.pdf'
    extensions['log'] = '.log'
    extensions['filewrite'] = ''
    extensions['saving_dta'] = '.dta'
    return extensions
### @end


### @export "program_def"
def program_def():
    "Defines Stata program"
    program = (lineStart + abbrev_key('pr', 'program') + restOfLine +
               originalTextFor(SkipTo(lineStart +
                                      Keyword('end'), include=False)) +
               Keyword('end') + restOfLine)
    return program
### @end


### @export "prepend"
def prepend():
    " Defines Stata's prepending commands "
    qui = abbrev_key('qui', 'quietly') + Optional(':')
    noi = abbrev_key('n', 'noisily') + Optional(':')
    cap = abbrev_key('cap', 'capture') + Optional(':')
    prepend = Suppress(ZeroOrMore(qui | noi | cap))
    return prepend
### @end


### @export "abbrev_key"
def abbrev_key(abbrev, long):
    """
    Helper function that allows to specify all possible
    pyparsing Keywords for Stata commands by just stating
    the shortest and longest expression.
    """
    abbrev_list = []
    for i in range(len(long)-len(abbrev)):
        abbrev_list.append(Keyword(long[:len(abbrev)+i]))
    abbrev_list.append(Keyword(long))
    abbrevs = Or(abbrev_list)
    return abbrevs
### @end


### @export "strip_white"
def strip_white(dofile,rmblanklines=False):
    """
    Helper function that converts line endings to \n
    and strips whitespace left and right. Optionally
    removes blank lines as well.
    """
    dofile = re.sub('\015\012','\n',dofile)
    dofile = re.sub('\015|\012','\n',dofile)
    # make sure that regardless of platform that file was 
    # written on, the end-of-line is a \n and nothing else 
    dofile = re.sub('[ \t]*\n[ \t]*','\n',dofile)
    # strip trailing and leading whitespace, necessary for pyparsing
    if rmblanklines:
        dofile = re.sub('\n*\n','\n',dofile)
    return dofile
### @end


### @export "strip_stata_quotes"
def strip_stata_quotes(in_string):
    """
    Helper function that removes Stata-type quotes:
      - double quote, double quote + end of line
      - compound double quote, cmpd dbl quote + eol
    """
    out_string = in_string.strip()
    if out_string[0]=='"' and out_string[-1]=='"':
        out_string = out_string[1:-1]
    elif out_string[0]=='"':
        out_string = out_string[1:-1]
    elif out_string[0:1]=='`"' and out_string[-2:-1]==""""'""":
        out_string = out_string[2:-2]
    elif out_string[0:1]=='`"' :
        out_string = out_string[2:]
    return out_string
### @end